# This file is part of Scaffold
#
# Scaffold is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2019 Ledger SAS, written by Olivier Hériveaux


from time import sleep
from typing import Any, Optional
from scaffold import Scaffold


class NACKError(Exception):
    """
    This error is thrown when a STM32 devices responds with NACK byte to a
    command.
    """

    def __init__(self, tag: Any = None):
        super().__init__("Device returned NACK to a command")
        self.tag = tag


class MemorySection:
    """Describes a memory section of a device."""

    def __init__(self, start: int, end: int):
        """
        :param start: First address of the section.
        :param end: End address of the section (excluded from section, address
            of next section).
        """
        if end < start:
            raise ValueError("Invalid section addresses")
        self.start = start
        self.end = end

    @property
    def size(self):
        """Size in bytes of the section."""
        return self.end - self.start


class STM32Device:
    """Possible name and product ID tuple."""

    def __init__(
        self,
        name: str,
        pid: int,
        memory_mapping: dict[str, MemorySection],
        offset_rdp: int = 0,
    ):
        self.name = name
        self.pid = pid
        self.memory_mapping = memory_mapping
        self.offset_rdp = offset_rdp


class STM32:
    """
    Class for instrumenting STM32 devices using Scaffold board and API. The
    following Scaffold IOs are used:

    - D0: STM32 UART MCU RX pin, Scaffold UART TX
    - D1: STM32 UART MCU TX pin, Scaffold UART RX
    - D2: STM32 NRST pin for reset
    - D6: STM32 BOOT0 pin
    - D7: STM32 BOOT1 pin

    The uart0 peripheral of Scaffold board is used for serial communication
    with the ST bootloader.

    This class can communicate with the ST bootloader via USART1. This allows
    programming the Flash memory and then execute the loaded code.
    """

    # Acknoledge byte returned by STM32 bootloader.
    ACK = 0x79
    # Error byte returned by STM32 bootloader.
    NACK = 0x1F

    # Memory layout for STM32F20xxx
    map_f20xxx = {
        "option_bytes": MemorySection(0x1FFFC000, 0x1FFFC008),
        "system": MemorySection(0x1FFF0000, 0x1FFF7A10),
        "flash": MemorySection(0x08000000, 0x08100000),
    }

    # Memory layout for STM32F100xx written by Hugo LE GUEN, intern at IETR
    map_f100xx = {
        "option_bytes": MemorySection(0x1FFFF800, 0x1FFFF80F),
        "system": MemorySection(0x1FFFF000, 0x1FFFF7FF),
        "flash": MemorySection(0x08000000, 0x0801FFFF),
    }

    # Memory layout for STM32F4xxx
    map_f40xxx = map_f20xxx

    # Memory layout for STM32L431xx
    map_l431xx = {
        "option_bytes": MemorySection(0x1FFF7800, 0x1FFF7810),
        "otp": MemorySection(0x1FFF7000, 0x1FFF7400),
        "system": MemorySection(0x1FFF0000, 0x1FFF7000),
        "flash": MemorySection(0x08000000, 0x08040000),
    }

    # Memory layout for STM32U5A9
    map_u5 = {
        "otp": MemorySection(0x0BFA0000, 0x0BFA0200),
        "system": MemorySection(0x0BF90000, 0x0BF98000),
        "flash": MemorySection(0x08000000, 0x08080000),
    }

    # See AN2606 for PID values
    PIDS = [
        STM32Device("STM32F05xxx", 0x440, {}),
        STM32Device("STM32F030x8", 0x440, {}),
        STM32Device("STM32F03xx4/6", 0x444, {}),
        STM32Device("STM32F100xx", 0x420, map_f100xx),
        STM32Device("STM32F2xxxx", 0x411, map_f20xxx, offset_rdp=1),
        STM32Device("STM32F40xxx/41xxx", 0x413, map_f40xxx, offset_rdp=1),
        STM32Device("STM32F42xxx/43xxx", 0x419, map_f40xxx, offset_rdp=1),
        STM32Device("STM32L43xxx/44xxx", 0x435, map_l431xx),
        STM32Device("STM32L45xxx/46xxx", 0x462, {}),
        STM32Device("STM32L47xxx/48xxx", 0x415, {}),
        STM32Device("STM32L496xx/4A6xx", 0x461, {}),
        STM32Device("STM32L4Rxx/4Sxx", 0x470, {}),
        STM32Device("STM32U59x/5Ax", 0x481, map_u5),
    ]

    def __init__(self, scaffold: Scaffold):
        """
        :param scaffold: An instance of :class:`scaffold.Scaffold` which will
            be configured and used to communicate with STM32 daughter board.
        """
        self.scaffold = scaffold
        self.scaffold.timeout = 1
        self.nrst = scaffold.d2
        self.uart = scaffold.uart0
        self.boot0 = scaffold.d6
        self.boot1 = scaffold.d7
        # Connect the UART peripheral to D0 and D1.
        self.uart.rx << scaffold.d1
        scaffold.d0 << self.uart.tx
        self.uart.baudrate = 115200
        # Instance of STM32Device, set when reading the device ID.
        self.device: Optional[STM32Device] = None

    def checksum(self, data: bytes) -> int:
        """
        Calculate the checksum of some data, according to the STM32
        bootloader protocol.

        :param data: Input bytes.
        :return: Checksum byte. This is the XOR of all input bytes.
        """
        result = 0x00
        for b in data:
            result ^= b
        return result

    def startup_bootloader(self):
        """
        Power-cycle and reset target device in bootloader mode (boot on System
        Memory) and initiate serial communication. The byte 0x7f is sent and
        the response byte 0x79 (ACK) is expected. If the device does not
        respond, a Timeout exception is thrown by Scaffold. The device will not
        respond if it is locked in RDP2 state (Readout Protection level 2).
        """
        self.scaffold.power.dut = 0
        _ = self.boot0 << 1
        _ = self.boot1 << 0
        _ = self.nrst << 0
        sleep(0.1)
        self.scaffold.power.dut = 1
        sleep(0.1)
        _ = self.nrst << 1
        sleep(0.1)
        # Send 0x7f byte for initiating communication
        self.uart.flush()
        self.uart.transmit(b"\x7f")
        self.wait_ack()

    def startup_flash(self):
        """
        Power-cycle and reset target device and boot from user Flash memory.
        """
        self.scaffold.power.dut = 0
        _ = self.boot0 << 0
        _ = self.boot1 << 0
        _ = self.nrst << 0
        sleep(0.1)
        self.scaffold.power.dut = 1
        sleep(0.1)
        _ = self.nrst << 1
        sleep(0.1)

    def command(self, index: int) -> bytes:
        """
        Send a command and return the response.

        :param index: Command index.
        :return: Response bytes.
        """
        self.uart.transmit(bytes([index, 0xFF ^ index]))
        res = self.uart.receive(2)
        assert res[0] == self.ACK
        data = self.uart.receive(res[1] + 2)
        assert data[-1] == self.ACK
        return data[0:-1]

    def wait_ack(self, tag: Optional[Any] = None):
        """
        Wait for ACK byte.

        :param tag: Tag which is set when NACKError are thrown. Useful for
            error diagnostic.
        """
        b = self.uart.receive(1)[0]
        if b == self.NACK:
            raise NACKError(tag)
        if b != self.ACK:
            raise Exception(f"Received 0x{b:02x} byte instead of ACK or NACK.")

    def wait_ack_or_nack(self) -> bool:
        """
        Wait for ACK or NACK byte.

        :return: True if ACK has been received, False if NACK has been
            received.
        """
        b = self.uart.receive(1)[0]
        assert b in (self.ACK, self.NACK)
        return b == self.ACK

    def get(self) -> bytes:
        """
        Execute the Get command of the bootloader, which returns the version
        and the supported commands.
        """
        response = self.command(0x00)
        return response

    def get_id(self) -> int:
        """
        Execute the Get ID command. The result is interpreted and the class
        will try to find information if the ID matches a known device.
        """
        self.device = None
        response = self.command(0x02)
        pid = int.from_bytes(response, "big", signed=False)
        for dev in self.PIDS:
            if dev.pid == pid:
                self.device = dev
        return pid

    def get_version_and_read_protection_status(self) -> bytes:
        self.uart.transmit(b"\x01\xfe")
        response = self.uart.receive(5)
        assert response[0] == self.ACK
        assert response[-1] == self.ACK
        return response[1:-1]

    def read_memory(
        self, address: int, length: int, trigger: bool = False
    ) -> bytes:
        """
        Tries to read some memory from the device. If requested size is larger
        than 256 bytes, many Read Memory commands are sent.

        :param address: Memory address to be read.
        :param size: Number of bytes to be read.
        :param trigger: True to enable trigger on command transmission.
        """
        result = bytearray()
        remaining = length
        while remaining > 0:
            chunk_size = min(256, remaining)
            self.uart.transmit(b"\x11\xee", trigger=trigger)
            self.wait_ack()
            buf = bytearray(address.to_bytes(4, "big", signed=False))
            buf.append(self.checksum(buf))
            self.uart.transmit(bytes(buf))
            self.wait_ack()
            buf = bytearray()
            buf.append(chunk_size - 1)
            buf.append((chunk_size - 1) ^ 0xFF)
            self.uart.transmit(bytes(buf))
            self.wait_ack()
            result += self.uart.receive(chunk_size)
            remaining -= chunk_size
            address += chunk_size
        return result

    def write_memory(self, address: int, data: bytes, trigger: bool = False):
        """
        Write data to device memory. If target address is Flash memory, this
        function DOES NOT erase Flash memory prior to writing. If data size is
        larger than 256 bytes, many Write Memory commands are sent.

        :param address: Address.
        :param data: Data to be written. bytes or bytearray.
        :param trigger: True to enable trigger on each command transmission.
        """
        remaining = len(data)
        offset = 0
        while remaining > 0:
            chunk_size = min(256, remaining)
            self.uart.transmit(b"\x31\xce", trigger=trigger)
            self.wait_ack(0)
            buf = bytearray((address + offset).to_bytes(4, "big", signed=False))
            buf.append(self.checksum(buf))
            self.uart.transmit(bytes(buf))
            self.wait_ack(1)
            buf = bytearray()
            buf.append(chunk_size - 1)
            buf += data[offset : offset + chunk_size]
            buf.append(self.checksum(buf))
            self.uart.transmit(bytes(buf))
            self.wait_ack(2)
            offset += chunk_size
            remaining -= chunk_size

    def assert_device(self):
        """Raise a RuntimeError is device is unknown (None)."""
        if self.device is None:
            raise RuntimeError(
                "Unknown device or not discovered yet. Call get_id or set the "
                "device attribute."
            )

    def read_option_bytes(self) -> bytes:
        """
        Read the option bytes of the device. The method get_id must have been
        called previously for device identification.

        :return: Memory content of 'option_bytes' section.
        """
        self.assert_device()  # We need the memory mapping
        assert self.device is not None
        section = self.device.memory_mapping["option_bytes"]
        return self.read_memory(section.start, section.size)

    def readout_protect(self):
        """
        Execute the Readout Unprotect command.
        """
        self.uart.transmit(b"\x82\x7d", 1)
        self.wait_ack()
        self.wait_ack()

    def readout_unprotect(self):
        """
        Execute the Readout Unprotect command. If the device is locked, it will
        perform mass flash erase, which can be very very long.
        """
        self.uart.transmit(b"\x92\x6d", 1)
        self.wait_ack()
        # When the chip is in RDP1 it will perform mass flash erase. This can
        # take a lot of time, so we must change the timeout setting.
        previous_timeout = self.scaffold.timeout
        self.scaffold.timeout = 30
        try:
            self.wait_ack()
        finally:
            # Restore timeout setting, even if something bad happened!
            self.scaffold.timeout = previous_timeout

    def write_protect(self, sectors: list[int]):
        """
        Execute Write Protect command.

        :param sectors: List of sectors to be write protected.
        """
        if len(sectors) not in range(1, 0x100):
            raise ValueError("Invalid sector count")
        self.uart.transmit(b"\x63\x9c")
        self.wait_ack()
        buf = bytearray()
        buf.append(len(sectors) - 1)
        buf += bytes(sectors)
        buf.append(self.checksum(buf))
        self.uart.transmit(bytes(buf))
        self.wait_ack()

    def write_unprotect(self):
        """
        Execute Write Unprotect command.
        """
        self.uart.transmit(b"\x73\x8c")
        self.wait_ack()
        self.wait_ack()

    def extended_erase(self):
        """
        Execute the Extended Erase command to erase all the Flash memory of the
        device.
        """
        self.uart.transmit(b"\x44\xbb")
        self.wait_ack()
        buf = bytearray(b"\xff\xff")
        buf.append(self.checksum(buf))
        self.uart.transmit(bytes(buf), trigger=True)
        previous_timeout = self.scaffold.timeout
        self.scaffold.timeout = 30
        try:
            self.wait_ack()
        finally:
            self.scaffold.timeout = previous_timeout

    def global_erase(self):
        """
        This function works with a chip that has a bootloader version lower
        than 3.0 like STM32 F100xx.
        Unlike a chip that has a bootloader version higher or equal to 3.0,
        the command to erase the Flash memory is x43/xbc instead of x44/xbb.
        The function is based on the translation of the diagram in the
        documentation "USART protocol used in the STM32" in
        the category "Erase Memory command: host side".
        Written by Hugo LE GUEN, intern at IETR.
        """
        self.uart.transmit(b"\x43\xbc")
        self.wait_ack()
        self.uart.transmit(b"\xff")
        self.uart.transmit(b"\x00")
        self.wait_ack()

    def go(self, address: int, trigger: bool = False):
        """
        Execute the Go command.

        :param address: Jump address.
        :param trigger: True to enable trigger on command transmission.
        """
        self.uart.transmit(b"\x21\xde", trigger=trigger)
        self.wait_ack()
        buf = bytearray(address.to_bytes(4, "big", signed=False))
        buf.append(self.checksum(buf))
        self.uart.transmit(bytes(buf))
        self.wait_ack()



#!/usr/bin/python3

from time import sleep
from enum import Enum
from binascii import hexlify
from scaffold import Pull, IOMode, Scaffold  # , I2CTrigger
from typing import Optional, List, Union, cast
import crcmod
from Crypto.Hash import SHA256


class I2CTrigger(str, Enum):
    START = "a"
    END = "b"


base_crc16 = crcmod.mkCrcFun(0x18005, 0, rev=True)


def crc16(data):
    x = base_crc16(data)
    rev = int(f"{x:016b}"[::-1], 2)
    result = ((rev & 0xFF) << 8) + (rev >> 8)
    return result


def crc16b(data):
    return int.to_bytes(crc16(data), 2, "big")


def swi_bits(byte: int) -> bytearray:
    result = bytearray()
    for i in range(8):
        result.append(0b11111101 + (((byte >> i) & 1) << 1))
    return result


class CRCError(Exception):
    def __init__(self, data, expected, received):
        self.data = data
        self.expected = expected
        self.received = received

    def __str__(self):
        return "CRCError"


class ProtocolError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class ATECCInterface(Enum):
    SWI = 1
    I2C = 2


class ATECCRevision(Enum):
    ATECC508A = 0x5000
    ATECC608A = 0x6002
    ATECC608B = 0x6003

    @property
    def is_ATECC608(self):
        return self in [ATECCRevision.ATECC608A, ATECCRevision.ATECC608B]

    @property
    def is_ATECC508(self):
        return self in [ATECCRevision.ATECC508A]


class ATECCErrorCode(int, Enum):
    SUCCESS = 0x00
    CHECKMAC_OR_VERIFY_MISCOMPARE = 0x01
    PARSE_ERROR = 0x03
    ECC_FAULT = 0x05
    SELF_TEST_ERROR = 0x07  # ATECC608A-TNGTLS
    HEALTH_TEST_ERROR = 0x08  # ATECC608A-TNGTLS
    EXECUTION_ERROR = 0x0F
    AFTER_WAKE = 0x11
    WATCHDOG_ABOUT_TO_EXPIRE = 0xEE  # ATECC608A-TNGTLS
    CRC_OR_OTHER_COMMUNICATION_ERROR = 0xFF  # ATECC608A-TNGTLS


class ATECCError(Exception):
    def __init__(self, code):
        super().__init__()
        self.code = code

    def __str__(self):
        try:
            return ATECCErrorCode(self.code).name
        except ValueError as e:
            return f"Unknown ATECC error 0x{self.code:02x}"


class ATECCOpCode(Enum):
    """ATECC508 command OpCodes"""

    CHECK_MAC = 0x28
    COUNTER = 0x24
    DERIVE_KEY = 0x1C
    ECDH = 0x43
    GEN_DIG = 0x15
    GEN_KEY = 0x40
    HMAC = 0x11
    INFO = 0x30
    LOCK = 0x17
    MAC = 0x08
    NONCE = 0x16
    PAUSE = 0x01
    PRIV_WRITE = 0x46
    RANDOM = 0x1B
    READ = 0x02
    SIGN = 0x41
    SHA = 0x47
    UPDATE_EXTRA = 0x20
    VERIFY = 0x45
    WRITE = 0x12
    AES = 0x51  # ATECC608A


class ATECCWordAddress(Enum):
    """Possible word-address for ATECC508 I2C writes."""

    RESET = 0x00
    SLEEP = 0x01
    IDLE = 0x02
    COMMAND = 0x03


class ATECCZone(Enum):
    """
    ATECC memory zones.
    Encoded values correspond to what must be transmitted in the READ commands.
    """

    CONFIG = 0
    OTP = 1
    DATA = 2


class SlotConfig:
    """ATECC data slot configuration."""

    def __init__(self):
        self.from_bytes(b"\x00\x00")

    def from_bytes(self, buf, offset=0):
        """
        Deserialize data from an input buffer.
        :param data: Input buffer.
        :type data: bytes
        :param offset: Offset of the data in the buffer.
        :type offset: int
        :return: Offset after the read data.
        :rtype: int
        """
        self.raw = buf[offset : offset + 2]
        x = int.from_bytes(self.raw, "little")
        self.write_config = (x >> 12) & 0x0F
        self.write_key = (x >> 8) & 0x0F
        self.is_secret = bool(x & (1 << 7))
        self.encrypt_read = bool(x & (1 << 6))
        self.limited_use = bool(x & (1 << 5))
        self.no_mac = bool(x & (1 << 4))
        self.read_key = x & 0x0F
        self.__check()
        return offset + 2

    def to_bytes(self):
        """
        :return: Data serialized to bytes.
        :rtype: bytes
        """
        self.__check()
        value = (
            (self.write_config << 12)
            + (self.write_key << 8)
            + (int(self.is_secret) << 7)
            + (int(self.encrypt_read) << 6)
            + (int(self.limited_use) << 5)
            + (int(self.no_mac) << 4)
            + self.read_key
        )
        return value.to_bytes(2, "little")

    def to_yaml(self) -> dict:
        """Generate dict object for yaml serialization"""
        self.__check()
        return {
            "write-config": self.write_config,
            "write-key": self.write_key,
            "is-secret": self.is_secret,
            "encrypt-read": self.encrypt_read,
            "limited-use": self.limited_use,
            "no-mac": self.no_mac,
            "read-key": self.read_key,
        }

    def from_yaml(self, data: dict):
        """Load attributes from yaml loaded dict"""
        self.write_config = data["write-config"]
        self.write_key = data["write-key"]
        self.is_secret = data["is-secret"]
        self.encrypt_read = data["encrypt-read"]
        self.limited_use = data["limited-use"]
        self.no_mac = data["no-mac"]
        self.read_key = data["read-key"]
        self.__check()

    def __check(self):
        """Check attributes types and values"""
        assert self.write_config in range(16)  # 4 bits
        assert self.write_key in range(16)  # 4 bits
        assert type(self.is_secret) == bool
        assert type(self.encrypt_read) == bool
        assert type(self.limited_use) == bool
        assert type(self.no_mac) == bool
        assert self.read_key in range(16)  # 4 bits


class ATECCInfoMode(Enum):
    REVISION = 0
    KEYVALID = 1
    STATE = 2
    GPIO = 3
    VOLATILE_KEY_PERMIT = 4


class Info:
    def __init__(self, mode, data: bytes, param):
        self.mode = mode
        self.revision: Optional[ATECCRevision] = None
        self.raw = data
        self.param = param
        if data is not None:
            self.from_bytes(data)

    def from_bytes(self, data: bytes, offset: int = 0):
        """
        Deserialize data from an input buffer.

        :param data: Input buffer.
        :param offset: Offset of the data in the buffer.
        :return: Offset after the read data.
        """
        self.raw = data
        buf = bytes(data[offset : offset + 4])
        if self.mode == ATECCInfoMode.REVISION:
            self.revision = ATECCRevision(int.from_bytes(buf, "big"))
        if self.mode == ATECCInfoMode.STATE:
            self.tempkey_id = buf[0] & 0xF
            self.tempkey_sourceflag = bool((buf[0] >> 4) & 0x1)
            self.tempkey_gendigdata = bool((buf[0] >> 5) & 0x1)
            self.tempkey_genkeydata = bool((buf[0] >> 6) & 0x1)
            self.tempkey_nomacflag = bool((buf[0] >> 7) & 0x1)

            self.authvalid = bool((buf[1] >> 2) & 0x1)
            self.authkey = (buf[1] >> 3) & 0b1111
            self.tempkey_valid = bool((buf[1] >> 7) & 0x1)

    def __str__(self) -> str:
        ret = f"Info mode: {self.mode} : {self.raw.hex()}\n"
        if self.mode == ATECCInfoMode.REVISION:
            ret += f" revision: {self.revision}"
        elif self.mode == ATECCInfoMode.KEYVALID:
            zeros = b"\0" * 4
            ret += f" Key {self.param} is valid? {self.raw != zeros}"
        elif self.mode == ATECCInfoMode.STATE:
            ret += f" tempkey_id: {self.tempkey_id}\n"
            ret += f" tempkey_sourceflag: {self.tempkey_sourceflag}\n"
            ret += f" tempkey_gendigdata: {self.tempkey_gendigdata}\n"
            ret += f" tempkey_genkeydata: {self.tempkey_genkeydata}\n"
            ret += f" tempkey_nomacflag: {self.tempkey_nomacflag}\n"
            ret += f" authvalid: {self.authvalid}\n"
            ret += f" authkey: {self.authkey}\n"
            ret += f" tempkey_valid: {self.tempkey_valid}"
        return ret


class KeyConfig:
    """ATECC key configuration"""

    def __init__(self, from_bytes: bytes = b"\x00\x00"):
        self.from_bytes(from_bytes)

    def from_bytes(self, buf, offset=0):
        """
        Deserialize data from an input buffer.
        :param data: Input buffer.
        :type data: bytes
        :param offset: Offset of the data in the buffer.
        :type offset: int
        :return: Offset after the read data.
        :rtype: int
        """
        self.raw = buf[offset : offset + 2]
        x = int.from_bytes(self.raw, "little")
        self.x509_id = x >> 14
        assert ((x >> 13) & 1) == 0  # RFU bit
        self.intrusion_disable = bool(x & (1 << 12))
        self.auth_key = (x >> 8) & 0x0F
        self.req_auth = bool(x & (1 << 7))
        self.req_random = bool(x & (1 << 6))
        self.lockable = bool(x & (1 << 5))
        self.key_type = (x >> 2) & 0b111
        self.pub_info = bool(x & (1 << 1))
        self.private = bool(x & 1)
        self.__check()
        return offset + 2

    def to_bytes(self):
        """
        :return: Data serialized to bytes.
        :rtype: bytes
        """
        self.__check()
        value = (
            (self.x509_id << 14)
            + (int(self.intrusion_disable) << 12)
            + (self.auth_key << 8)
            + (int(self.req_auth) << 7)
            + (int(self.req_random) << 6)
            + (int(self.lockable) << 5)
            + (self.key_type << 2)
            + (int(self.pub_info) << 1)
            + int(self.private)
        )
        return value.to_bytes(2, "little")

    def from_yaml(self, data: dict):
        """Load attributes from yaml loaded dict"""
        self.x509_id = data["x509-id"]
        self.intrusion_disable = data["intrusion-disable"]
        self.auth_key = data["auth-key"]
        self.req_auth = data["req-auth"]
        self.req_random = data["req-random"]
        self.lockable = data["lockable"]
        self.key_type = data["key-type"]
        self.pub_info = data["pub-info"]
        self.private = data["private"]
        self.__check()

    def to_yaml(self) -> dict:
        """Generate dict object for yaml serialization"""
        self.__check()
        return {
            "x509-id": self.x509_id,
            "intrusion-disable": self.intrusion_disable,
            "auth-key": self.auth_key,
            "req-auth": self.req_auth,
            "req-random": self.req_random,
            "lockable": self.lockable,
            "key-type": self.key_type,
            "pub-info": self.pub_info,
            "private": self.private,
        }

    def __check(self):
        """Check attributes types and values"""
        assert self.x509_id in range(4)  # 2 bits
        assert type(self.intrusion_disable) == bool
        assert self.auth_key in range(16)  # 4 bits
        assert type(self.req_auth) == bool
        assert type(self.req_random) == bool
        assert type(self.lockable) == bool
        assert self.key_type in range(8)  # 3 bits
        assert type(self.pub_info) == bool
        assert type(self.private) == bool


class Config:
    """
    Image of the configuration of the device (EEPROM configuration memory)
    """

    def __init__(self, from_bytes=None):
        self.serial = None
        self.revision = None
        self.atecc_revision = None
        self.internal_13 = None
        self.aes_enable = None
        self.internal_14 = None
        self.i2c_enable = None
        self.internal_15 = None
        self.i2c_address = None
        self.internal_17 = None
        self.otp_mode = None
        self.count_match_enable = None
        self.count_match_key = None
        self.clock_divider = None
        self.watchdog_duration = None
        self.ttl_enable = None
        self.selector_mode = None
        self.slot_config = list(SlotConfig() for _ in range(16))
        self.counters: List[Optional[bytes], Optional[bytes]] = [None, None]
        self.last_key_use = None
        self.user_extra = None
        self.selector = None
        self.lock_value = None
        self.lock_config = None
        self.slot_locked: List[Optional[bool]] = list(None for _ in range(16))
        self.chip_options = None
        self.x509_format = None
        self.key_config = list(KeyConfig() for _ in range(16))
        self.raw = None
        if from_bytes is not None:
            self.from_bytes(from_bytes)

    def from_bytes(self, buf: bytes, offset: int = 0) -> int:
        """
        Deserialize data from an input buffer.

        :param buf: Input buffer.
        :param offset: Offset of the data in the buffer.
        :return: Offset after the read data.
        """
        buf = bytes(buf[offset : offset + 4 * 32])
        self.raw = buf
        self.serial = bytes(buf[0:4] + buf[8:13])
        self.revision = buf[4:8]
        revision = int.from_bytes(buf[4:8], "big")
        # For all versions of ATECC508A, 3rd byte will always be 0x50
        # For all versions of ATECC608A/B, 3rd byte will always be 0x60
        # For ATECC608A devices, 4th byte will be 0x02
        # For ATECC608B devices, 4th byte will be 0x03
        self.atecc_revision = ATECCRevision(revision)
        # Byte 13 'reserved' for ATECC508A, but AES enable for ATECC608A
        x = buf[13]
        if self.atecc_revision.is_ATECC608:
            self.internal_13 = x & 0xFE
            self.aes_enable = bool(x & 1)
        else:
            self.internal_13 = x
        # I2C enable
        x = buf[14]
        self.internal_14 = x & 0xFE
        self.i2c_enable = bool(x & 1)
        self.internal_15 = buf[15]
        self.i2c_address = buf[16]
        self.internal_17 = buf[17]
        assert self.internal_17 == 0
        # OTP mode (508A) or CountMatch (608A/B)
        x = buf[18]
        if self.atecc_revision.is_ATECC508:
            self.otp_mode = x
        elif self.atecc_revision.is_ATECC608:
            # CountMatch for ATECC608A
            self.count_match_enable = bool(x & 1)
            assert (x & 0b1110) == 0
            self.count_match_key = (x >> 4) & 0b1111
        # Chip mode
        x = buf[19]
        if self.atecc_revision.is_ATECC508:
            assert (x & 0b11111000) == 0
            self.selector_mode = x & 1
        elif self.atecc_revision.is_ATECC608:
            self.i2c_address_user_extra = bool(x & 1)
            self.clock_divider = (x >> 3) & 0b11111
        self.ttl_enable = bool((x >> 1) & 1)
        self.watchdog_duration = [1.3, 10][(x >> 2) & 1]
        # Slot config
        for i in range(16):
            self.slot_config[i].from_bytes(buf, 20 + i * 2)
        self.counters[0] = buf[52:60]
        self.counters[1] = buf[60:68]
        if self.atecc_revision.is_ATECC508:
            self.last_key_use = buf[68:84]
        elif self.atecc_revision.is_ATECC608:
            x = buf[68]
            self.use_lock_enable = (x & 0b1111) == 0x0A
            self.use_lock_key = (x >> 4) & 0b1111
            x = buf[69]
            self.volatile_key_permit_slot = x & 0b1111
            assert x & 0b1110000 == 0
            self.volatile_key_permit_enable = ((x >> 7) & 0b1) == 1
            self.secure_boot = buf[70:72]
            self.kdf_iv_loc = buf[72]
            self.kdf_iv_str = buf[73:75]
            assert buf[75:84] == bytes(84 - 75)
        self.user_extra = buf[84]
        self.selector = buf[85]
        # Lock value
        x = buf[86]
        # assert x in (0x55, 0x00)
        self.lock_value = x
        # Lock config
        x = buf[87]
        # assert x in (0x55, 0x00)
        self.lock_config = x
        # Slot locked
        x = int.from_bytes(buf[88:90], "little")
        for i in range(16):
            self.slot_locked[i] = (x & 1) == 0
            x >>= 1
        if self.atecc_revision.is_ATECC508:
            # For rev 508: Reserved, shall always be zero
            assert buf[90] == 0
            assert buf[91] == 0
        elif self.atecc_revision.is_ATECC608:
            self.chip_options = buf[90:92]
            # bits 3-7 must be zero
            assert (self.chip_options[0] & 0b11111000) == 0
        # X509 format
        self.x509_format = buf[92:96]
        # Key config
        for i in range(16):
            self.key_config[i].from_bytes(buf, 96 + i * 2)
        self.__check()
        return offset + 4 * 32

    def to_bytes(self):
        """
        :return: Data serialized to bytes.
        :rtype: bytes
        """
        self.__check()
        res = bytearray(self.serial[0:4] + self.revision + self.serial[4:9])
        if self.atecc_revision.is_ATECC608:
            res.append(int(self.aes_enable) + self.internal_13)
        else:
            res.append(self.internal_13)
        res.append(int(self.i2c_enable) + self.internal_14)
        res.append(self.internal_15)  # Reserved
        res.append(self.i2c_address)
        res.append(self.internal_17)  # Reserved
        if self.atecc_revision.is_ATECC508:
            res.append(self.otp_mode)
        elif self.atecc_revision.is_ATECC608:
            res.append(
                (1 if self.count_match_enable else 0) + (self.count_match_key << 4)
            )
        x = ((1 if self.ttl_enable else 0) << 1) + {1.3: 0, 10: 1 << 2}[
            self.watchdog_duration
        ]
        if self.atecc_revision.is_ATECC508:
            x += self.selector_mode
        elif self.atecc_revision.is_ATECC608:
            x += (1 if self.i2c_address_user_extra else 0) + (self.clock_divider << 3)
        res.append(x)
        for c in self.slot_config:
            res += c.to_bytes()
        res += self.counters[0]
        res += self.counters[1]
        if self.atecc_revision.is_ATECC508:
            res += self.last_key_use
        elif self.atecc_revision.is_ATECC608:
            res.append(0x0A if self.use_lock_enable else 0 + (self.use_lock_key << 4))
            res.append(
                self.volatile_key_permit_slot
                + ((1 if self.volatile_key_permit_enable else 0) << 7)
            )
            res += self.secure_boot
            res.append(self.kdf_iv_loc)
            res += self.kdf_iv_str
            res += bytes(9)
        res.append(self.user_extra)
        res.append(self.selector)
        res.append(self.lock_value)
        res.append(self.lock_config)
        x = 0
        for i in range(16):
            if not self.slot_locked[i]:
                x += 1 << i
        res += x.to_bytes(2, "big")
        if self.atecc_revision.is_ATECC508:
            res.append(0)
            res.append(0)
        elif self.atecc_revision.is_ATECC608:
            res += self.chip_options
        res += self.x509_format
        for c in self.key_config:
            res += c.to_bytes()
        return res

    def from_yaml(self, data: dict):
        """Load attributes from yaml loaded dict"""
        self.serial = bytes.fromhex(data["serial"])
        self.revision = bytes.fromhex(data["revision"])
        revision = int.from_bytes(self.revision[4:8], "big")
        self.atecc_revision = ATECCRevision(revision)
        if self.atecc_revision.is_ATECC608:
            raise RuntimeError(
                "Sorry, YAML import/export has not been updated for ATECC608"
            )
        self.internal_13 = data.get("internal-13", 0)
        self.aes_enable = data["aes-enable"]
        self.internal_14 = data.get("internal-14", 0)
        self.i2c_enable = data["i2c-enable"]
        self.i2c_address = data["i2c-address"]
        self.otp_mode = data["otp-mode"]
        self.watchdog_duration = data["watchdog-duration"]
        self.ttl_enable = data["ttl-enable"]
        self.selector_mode = data["selector-mode"]
        self.slot_config = []
        for i in range(16):
            c = SlotConfig()
            c.from_yaml(data["slot-config"][i])
            self.slot_config.append(c)
        self.counters = [bytes.fromhex(x) for x in data["counters"]]
        self.last_key_use = bytes.fromhex(data["last-key-use"])
        self.user_extra = data["user-extra"]
        self.selector = data["selector"]
        self.lock_value = data["lock-value"]
        self.lock_config = data["lock-config"]
        self.slot_locked = data["slot-locked"]
        self.x509_format = bytes.fromhex(data["x509-format"])
        self.key_config = []
        for i in range(16):
            c = KeyConfig()
            c.from_yaml(data["key-config"][i])
            self.key_config.append(c)
        self.__check()

    def to_yaml(self) -> dict:
        if self.atecc_revision.is_ATECC608:
            raise RuntimeError(
                "Sorry, YAML import/export has not been updated for ATECC608"
            )
        """Generate dict object for yaml serialization"""
        self.__check()
        res = {
            "serial": self.serial.hex(),
            "revision": self.revision.hex(),
            "aes-enable": self.aes_enable,
            "i2c-enable": self.i2c_enable,
            "i2c-address": self.i2c_address,
            "otp-mode": self.otp_mode,
            "watchdog-duration": self.watchdog_duration,
            "ttl-enable": self.ttl_enable,
            "selector-mode": self.selector_mode,
            "slot-config": [c.to_yaml() for c in self.slot_config],
            "counters": [c.hex() for c in self.counters],
            "last-key-use": self.last_key_use.hex(),
            "user-extra": self.user_extra,
            "selector": self.selector,
            "lock-value": self.lock_value,
            "lock-config": self.lock_config,
            "slot-locked": self.slot_locked,
            "x509-format": self.x509_format.hex(),
            "key-config": [c.to_yaml() for c in self.key_config],
        }
        if self.internal_13 != 0:
            res["internal-13"] = self.internal_13
        if self.internal_14 != 0:
            res["internal-14"] = self.internal_14
        return res

    def __check(self):
        """Check attributes types and values"""
        assert type(self.serial) == bytes
        assert len(self.serial) == 9
        assert type(self.revision) == bytes
        assert len(self.revision) == 4
        if self.atecc_revision.is_ATECC608:
            assert type(self.aes_enable) == bool
        assert type(self.i2c_enable) == bool
        assert self.i2c_address in range(0, 0x100)
        if self.atecc_revision.is_ATECC508:
            assert self.otp_mode in (0xAA, 0x55)
        if self.atecc_revision.is_ATECC508:
            assert self.selector_mode in range(0, 2)
        elif self.atecc_revision.is_ATECC608:
            assert type(self.i2c_address_user_extra) == bool
        assert type(self.ttl_enable) == bool
        if self.atecc_revision.is_ATECC608:
            assert self.clock_divider in [0x00, 0x0D, 0x05]
        assert self.watchdog_duration in [1.3, 10]
        assert len(self.slot_config) == 16
        for c in self.slot_config:
            assert isinstance(c, SlotConfig)
        assert len(self.counters) == 2
        for c in self.counters:
            assert len(c) == 8
        if self.atecc_revision.is_ATECC508:
            assert len(self.last_key_use) == 16
        elif self.atecc_revision.is_ATECC608:
            assert self.secure_boot[0] & 0b11100100 == 0
        assert self.user_extra in range(0, 0x100)
        assert self.selector in range(0, 0x100)
        assert self.lock_value in (0, 0x55)
        assert self.lock_config in (0, 0x55)
        assert len(self.slot_locked) == 16
        for x in self.slot_locked:
            assert type(x) == bool
        assert type(self.x509_format) == bytes
        assert len(self.x509_format) == 4
        assert len(self.key_config) == 16
        for c in self.key_config:
            assert isinstance(c, KeyConfig)

    def print_config(self):
        def yes_no(x) -> str:
            return "yes" if x else "no"

        """ Print given configuration """
        if self.raw is not None:
            print("Raw:")
            for i in range(4):
                print("  " + hexlify(self.raw[i * 32 : (i + 1) * 32]).decode())

        print("Serial: 0x" + hexlify(self.serial).decode())
        print(
            "Revision: 0x" + hexlify(self.revision).decode(), f"({self.atecc_revision})"
        )
        if self.atecc_revision.is_ATECC608:
            print("AES enable: " + ["no", "yes"][int(self.aes_enable)])
        if self.internal_13 != 0:
            print(f"Internal 13: {self.internal_13}")
        print("I2C enable: " + ["no", "yes"][int(self.i2c_enable)])
        if self.internal_14 != 0:
            print(f"Internal 14: {self.internal_14}")
        print(f"I2C address: 0x{self.i2c_address:02x}")
        x = self.otp_mode
        s = {0xAA: "read-only", 0x55: "consumption"}.get(x, "unknown")
        if self.atecc_revision == ATECCRevision.ATECC508A:
            print(f"OTP mode: 0x{x:02x} ({s})")
        elif self.atecc_revision.is_ATECC608:
            print(f"CountMatch enabled: {self.count_match_enable}")
            print(f"CountMatch key: {self.count_match_key}")
        print(f"Watchdog duration: {self.watchdog_duration} s")
        print(
            "TTL enable: " + ["fixed reference", "vcc referenced"][int(self.ttl_enable)]
        )
        print(f"Selector mode: {self.selector_mode}")
        if self.atecc_revision.is_ATECC608:
            print(f"Clock divider: {self.clock_divider}")
        print("Slot config:")
        for i, sc in enumerate(self.slot_config):
            print(f"  Slot {i}:")
            print("    Raw: 0x" + hexlify(sc.raw).decode())
            x = sc.write_config
            print(f"    Write config: 0b{x:04b}")
            if x == 0:
                s = "always"
            elif x == 1:
                s = "pub invalid"
            elif (x & 0b1110) == 0b0010:
                s = "never"
            elif (x & 0b1100) == 0b1000:
                s = "never"
            elif (x & 0b0100) == 0b0100:
                s = "encrypt"
            else:
                s = "?"
            print(f"      write: {s}")
            print(f"    Write key: {sc.write_key}")
            print(f"    Read key: {sc.read_key}")
            print(f"    Is secret: " + ["no", "yes"][int(sc.is_secret)])
            print(f"    Encrypt read: " + ["no", "yes"][int(sc.encrypt_read)])
            print(f"    Limited use: " + ["no", "yes"][int(sc.limited_use)])
            print(f"    No MAC: " + ["no", "yes"][int(sc.no_mac)])
        print("Counters:")
        for i, c in enumerate(self.counters):
            print(f"  {i}: 0x" + hexlify(c).decode())
        if self.atecc_revision.is_ATECC508:
            print("Last key use: " + hexlify(self.last_key_use).decode())
        elif self.atecc_revision.is_ATECC608:
            print("Volatile Key Permit Enabled:", self.volatile_key_permit_enable)
            print("Volatile Key Permit Slot:", self.volatile_key_permit_slot)
            mode = self.secure_boot[0] & 0b11
            if mode == 0b00:
                print(f"Secure boot mode {mode:02b}: Disabled")
            elif mode == 0b01:
                print(f"Secure boot mode {mode:02b}: Full Secure boot (FullBoth)")
            elif mode == 0b10:
                print(f"Secure boot mode {mode:02b}: Stored Secure Boot (FullSig)")
            elif mode == 0b11:
                print(f"Secure boot mode {mode:02b}: Stored secure Boot (FullDig)")
            print("KDF IV location:", self.kdf_iv_loc)
            print(f"KDF IV string: 0x{self.kdf_iv_str.hex()}")
        print(f"User extra: 0x{self.user_extra:02x}")
        if self.atecc_revision.is_ATECC508:
            print(f"Selector: 0x{self.selector:02x}")
        elif self.atecc_revision.is_ATECC608:
            print(f"User extra i2c address: 0x{self.selector:02x}")
        x = self.lock_value
        s = {0x55: "unlocked", 0x00: "locked"}.get(x, "?")
        print(f"Lock value: 0x{x:02x} ({s})")
        x = self.lock_config
        s = {0x55: "unlocked", 0x00: "locked"}.get(x, "?")
        print(f"Lock config: 0x{x:02x} ({s})")
        print("Slot locked:")
        for i, sl in enumerate(self.slot_locked):
            print(f"  {i}: " + ["no", "yes"][int(sl)])
        if self.atecc_revision.is_ATECC608:
            co = self.chip_options
            print(f"Chip Options: 0x{co.hex()}")
            print(f"   Power on Self Test: {(co[0] & 0b1) != 0}")
            print(f"   IO Protection Key Enabled: {(co[0] & 0b10) != 0}")
            print(f"   KDF AES Enabled: {(co[0] & 0b100) != 0}")
            print(f"   ECDSH Protection bits: 0b{co[1] & 0b11:02b}")
            print(f"   KDF Protection bits: 0b{(co[1] >> 2) & 0b11:02b}")
            print(f"   IO Protection Key: {(co[1] >> 4) & 0b1111}")
        print(f"X509 format: 0x{self.x509_format.hex()}")
        print("Key config:")
        for i, kc in enumerate(self.key_config):
            print(f"  Key {i}:")
            if kc.raw is not None:
                print("    Raw: 0x" + hexlify(kc.raw).decode())
            print(f"    X509 ID: {kc.x509_id}")
            print("    Intrusion disable: " + yes_no(kc.intrusion_disable))
            print(f"    Auth key: {kc.auth_key}")
            print("    Req auth: " + yes_no(kc.req_auth))
            print("    Req random: " + yes_no(kc.req_random))
            print("    Lockable: " + yes_no(kc.lockable))
            if self.revision[2] > 0x05:
                d = {4: "P256 NIST ECC key", 6: "AES key", 7: "SHA key or other data"}
            else:
                d = {4: "P256 NIST ECC key", 7: "Not an ECC key"}
            print("    Key type: " + d.get(kc.key_type, f"0x{kc.key_type}"))
            print("    Pub info: " + yes_no(kc.pub_info))
            print("    Private: " + yes_no(kc.private))


class ATECCIOFlag(Enum):
    COMMAND = 0x77
    TRANSMIT = 0x88
    IDLE = 0xBB
    SLEEP = 0xCC


class ATECC:
    """Factory default I2C address for ATECC508."""

    DEFAULT_ADDRESS = 0xC0
    SWI_TX_BAUDRATE = 1 / 4.34e-6

    def __init__(self, scaffold: Scaffold, interface=ATECCInterface.I2C):
        self.interface = interface
        self.scaffold = scaffold
        self.sda = scaffold.d0
        self.scl = scaffold.d1
        self.sda.pull = Pull.UP
        if interface == ATECCInterface.I2C:
            self.sda.mode = IOMode.AUTO
            self.i2c = scaffold.i2cs[0]
            # ATECC does not use clock stretchin.
            # Disabling it allows saving a pull-up resistor on SCL.
            self.i2c.clock_stretching = 0
            self.sda << self.i2c.sda_out
            self.scl << self.i2c.scl_out
            self.i2c.sda_in << self.sda
            self.i2c.scl_in << self.scl
            self.i2c.frequency = 500e3
            self.i2c.address = self.DEFAULT_ADDRESS
        elif interface == ATECCInterface.SWI:
            self.sda.mode = IOMode.OPEN_DRAIN
            self.uart = scaffold.uart0
            self.uart.baudrate = self.SWI_TX_BAUDRATE
            self.sda << self.uart.tx
            self.sda >> self.uart.rx
        self.serial = None  # Set when calling read_serial or read_config
        self.temp_key: Optional[bytes] = None  # Set when calling nonce

    @property
    def address(self):
        """I2C address of the device. Default is DEFAULT_ADDRESS."""
        if self.interface != ATECCInterface.I2C:
            raise RuntimeError("interface is not I2C")
        return self.i2c.address

    @address.setter
    def address(self, value):
        if self.interface != ATECCInterface.I2C:
            raise RuntimeError("interface is not I2C")
        self.i2c.address = value

    def power_cycle(self):
        self.scaffold.power.restart_dut(toff=0.05, ton=0.05)

    def wake_up(self):
        self.sda << 0
        sleep(2e-4)
        self.sda << 1
        sleep(2e-3)
        if self.interface == ATECCInterface.I2C:
            self.sda << self.i2c.sda_out
        elif self.interface == ATECCInterface.SWI:
            self.sda << self.uart.tx

    def __swi_receive(self, n):
        buf = self.uart.receive(n * 8)
        result = bytearray()
        for i in range(n):
            byte = 0
            for j in range(8):
                b = buf[i * 8 + j]
                if b == 0xFF:
                    byte += 1 << j
                elif b != 0xFB:
                    raise RuntimeError("Transmission error")
            result.append(byte)
        return result

    def __read_output_group(self):
        if self.interface == ATECCInterface.I2C:
            response = self.i2c.read(1)
        elif self.interface == ATECCInterface.SWI:
            # We must drop first 8 bits
            self.uart.receive(8)
            response = self.__swi_receive(1)
        if response[0] < 4:
            raise ProtocolError("I/O group is too small")
        if self.interface == ATECCInterface.I2C:
            response = response + self.i2c.read(response[0] - 1)
        elif self.interface == ATECCInterface.SWI:
            response += self.__swi_receive(response[0] - 1)
        # Check response checksum
        expected_crc = crc16b(response[:-2])
        received_crc = response[-2:]
        if crc16b(response[:-2]) != response[-2:]:
            raise CRCError(response, received_crc, expected_crc)
        # Return result without length and CRC
        return response[1:-2]

    def command(
        self,
        opcode: Union[int, ATECCOpCode],
        p1: int = 0,
        p2: int = 0,
        data: bytes = b"",
        trigger: Union[bool, int, I2CTrigger] = 0,
        wait: Optional[float] = 0.0,
        skip_response: bool = False,
    ) -> Optional[bytes]:
        """
        Execute a command on the ATECC.

        :param opcode: Command code.
        :param p1: First command parameter. 8-bits.
        :param p2: Second command parameter. 16-bits.
        :param data: Command data.
        :param trigger: Enable or disable trigger signal for this command.
        :param wait: Wait time after sending the command.
        :param skip_response: If enabled, don't fetch the response, only send the command.
        :return: Command response.
        """
        if isinstance(opcode, ATECCOpCode):
            opcode = opcode.value
        buf = bytearray()
        if self.interface == ATECCInterface.I2C:
            buf.append(ATECCWordAddress.COMMAND.value)
        elif self.interface == ATECCInterface.SWI:
            buf.append(ATECCIOFlag.COMMAND.value)
        buf.append(7 + len(data))
        buf.append(opcode)
        buf.append(p1)
        buf += p2.to_bytes(2, "little")
        buf += data
        buf += crc16b(buf[1:])
        if self.interface == ATECCInterface.I2C:
            _trigger = trigger.value if isinstance(trigger , I2CTrigger) else trigger
            self.i2c.write(buf, trigger=_trigger)
            if wait is not None:
                sleep(wait)
            else:
                # Default wait delay
                sleep(15e-3)
            if skip_response:
                return
        elif self.interface == ATECCInterface.SWI:
            bits = bytearray()
            for b in buf:
                bits += swi_bits(b)
            self.uart.transmit(bits, trigger=trigger)
            # For SWI, we must send a TRANSMIT byte to ask the device to
            # respond. This can be done only after processing is done, otherwise
            # the TRANSMIT byte is ignored. Therefore, it is mandatory to wait.
            if wait:
                sleep(wait)
            else:
                # Default wait delay
                sleep(15e-3)
            if skip_response:
                return
            self.uart.flush()
            self.uart.transmit(swi_bits(ATECCIOFlag.TRANSMIT.value))
        g = self.__read_output_group()
        if len(g) == 1:
            # Status
            if g[0] != 0:
                # Command is not successfull
                raise ATECCError(g[0])
            return b""
        else:
            return g

    def reset(self):
        """Reset the address counter. Next I2C will start with the beginning of IO buffer."""
        if self.interface == ATECCInterface.I2C:
            self.i2c.write(bytes((ATECCWordAddress.RESET.value,)))
        elif self.interface == ATECCInterface.SWI:
            pass

    def idle(self):
        """Put the device into IDLE mode."""
        if self.interface == ATECCInterface.I2C:
            self.i2c.write(bytes((ATECCWordAddress.IDLE.value,)))
        elif self.interface == ATECCInterface.SWI:
            self.uart.transmit(swi_bits(ATECCIOFlag.IDLE.value))

    def info(
        self, mode: ATECCInfoMode = ATECCInfoMode.REVISION, param=0, trigger=0
    ) -> Info:
        """
        Query info from device.
        :return: Device info bytes.
        :rtype: bytes
        """
        info = self.command(ATECCOpCode.INFO, mode.value, param, trigger=trigger)
        return Info(mode, info, param)

    def state(self, trigger=0):
        return self.info(ATECCInfoMode.STATE, trigger=trigger)

    def read(
        self,
        zone=ATECCZone.DATA,
        size=32,
        block=0,
        slot=None,
        offset=0,
        trigger: Union[bool, int, I2CTrigger] = 0,
    ):
        """
        Read memory from the device.

        :param zone: Memory zone.
        :type zone: ATECCZone
        :param slot: Slot index, for data memory only. None for other zones.
        :type slot: int or None
        :param block: Block index in the selected slot or zone.
        :type block: int
        :param offset: Read/write offset
        :type offset: int
        :param size: 4 or 32.
        :type size: int
        :return: Memory content.
        :rtype: bytes
        """
        if size not in (4, 32):
            raise ValueError("size must be 4 or 32.")
        return self.command(
            ATECCOpCode.READ,
            ({4: 0, 32: 1}[size] << 7) | zone.value,
            self.__make_addr(zone, block, slot, offset),
            trigger=trigger,
        )

    def encrypted_read(self, slot, key_id, key, trigger=0):
        self.nonce()
        self.gen_dig(key_id, key)
        encrypted_data = self.read(
            zone=ATECCZone.DATA, size=32, slot=slot, block=0, trigger=trigger
        )
        # Decrypt the data with temp key (XOR)
        assert self.temp_key is not None
        assert encrypted_data is not None
        return bytes([a ^ b for a, b in zip(self.temp_key, encrypted_data)])

    def encrypted_write(self, slot, key_id, key, data, trigger=0):
        self.nonce()
        self.gen_dig(key_id, key)
        # Encrypt the data with temp key (XOR)
        encrypted_data = bytes([a ^ b for a, b in zip(self.temp_key, data)])
        # SHA-256(TempKey, Opcode, Param1, Param2, SN[8], SN[0:1], <25 bytes of zeros>, PlainTextData)
        param1 = ({4: 0, 32: 1}[len(data)] << 7) | ATECCZone.DATA.value
        param2 = self.__make_addr(ATECCZone.DATA, 0, slot, 0).to_bytes(2, "little")
        sn = self.serial
        sha = SHA256.new()
        sha.update(self.temp_key)
        sha.update(bytes([ATECCOpCode.WRITE.value, param1]))
        sha.update(param2)
        sha.update(sn[8:9] + sn[0:2] + b"\0" * 25 + data)
        mac = sha.digest()
        self.write(
            zone=ATECCZone.DATA,
            data=encrypted_data,
            mac=mac,
            slot=slot,
            block=0,
            trigger=trigger,
        )

    def __make_addr(self, zone, block, slot=None, offset=0):
        """
        Generate address for use in write and read commands.

        :param zone: Memory zone.
        :type zone: ATECCZone
        :param slot: Slot index, for data memory only. None for other zones.
        :type slot: int or None
        :param block: Block index in the selected slot or zone.
        :type block: int
        :param offset: Read/write offset
        :type offset: int
        :return: Encoded address.
        :rtype: int
        """
        # Check the parameters
        if offset not in range(8):
            raise IndexError("Invalid offset")
        if (zone in (ATECCZone.CONFIG, ATECCZone.OTP)) and (slot is not None):
            raise ValueError("Slot must be None for the selected zone")
        if zone == ATECCZone.DATA:
            if slot not in range(16):
                raise IndexError(f"Invalid slot {slot}. Should be less than 16.")
            if slot < 8:
                block_range = 2
            elif slot == 8:
                block_range = 8
            else:
                block_range = 4
            if block not in range(block_range):
                raise IndexError("Invalid block")
        # Encode address depending on the memory zone
        if zone == ATECCZone.CONFIG:
            return (block << 3) + offset
        elif zone == ATECCZone.OTP:
            return (block << 3) + offset
        else:
            return (block << 8) + (slot << 3) + offset

    def write(
        self,
        data: bytes,
        zone=ATECCZone.DATA,
        block=0,
        slot=None,
        offset=0,
        trigger=0,
        mac=None,
    ):
        """
        Write memory to the device.

        :param zone: Memory zone.
        :type zone: ATECCZone
        :param data: Data to be written. Size must be 4 or 32.
        :type data: bytes
        :param slot: Slot index, for data memory only. None for other zones.
        :type slot: int or None
        :param block: Block index in the selected slot or zone.
        :type block: int
        :param offset: Read/write offset
        :type offset: int
        """
        size = len(data)
        if size not in (4, 32):
            raise ValueError("size must be 4 or 32.")
        result = self.command(
            ATECCOpCode.WRITE,
            ({4: 0, 32: 1}[size] << 7) | zone.value,
            self.__make_addr(zone, block, slot, offset),
            data + (mac if mac is not None else b""),
            wait=0.1,
            trigger=trigger,
        )
        if result != b"":
            raise ProtocolError(
                "Received unexpected response 0x" + hexlify(result).decode()
            )

    def read_serial(self) -> bytes:
        """
        Read and save in `self.serial` the device serial number.

        :return: Device serial number
        """
        buf = self.read(ATECCZone.CONFIG, 32, 0)
        self.serial = buf[0:4] + buf[8:13]
        return self.serial

    def read_config(self):
        """
        Read all the configuration memory.
        :return: Parsed configuration memory.
        :rtype: Config
        """
        data = bytearray()
        for i in range(4):
            self.wake_up()
            data += self.read(ATECCZone.CONFIG, 32, i)
            self.idle()
        config = Config()
        config.from_bytes(data, 0)
        self.serial = config.serial
        return config

    def lock_config(self):
        """
        Lock config zone of the EEPROM memory.
        """
        # Bit 7 to ignore CRC content checking before lock.
        mode = 1 << 7
        self.command(ATECCOpCode.LOCK, mode, 0)

    def lock_data(self):
        """
        Lock data zone of the EEPROM memory.
        """
        # Bit 7 to ignore CRC content checking before lock.
        mode = (1 << 7) + 0b01
        self.command(ATECCOpCode.LOCK, mode, 0)

    def lock_slot(self, slot: int):
        """
        Lock individual data slot.
        :param slot: Slot number.
        """
        if slot not in range(0x10):
            raise ValueError(
                f"Incorrect value for slot number {slot}: should be less that {0x10}"
            )

        # Bit 7 to ignore CRC content checking before lock.
        mode = ((slot & 0b1111) << 2) + 0b10
        self.command(ATECCOpCode.LOCK, mode, 0)

    def gen_priv_key(self, slot: int, trigger=0):
        """
        Call the GenKey command to generate a private key in a slot.

        :param slot: Slot number.
        :param trigger: permits to generate a trigger on the command.
        """
        # Command execution is long. We must wait before querying the response.
        self.command(ATECCOpCode.GEN_KEY, 1 << 2, slot, trigger=trigger, wait=0.2)

    def gen_pub_key(self, slot, trigger=0):
        """
        Call the GenKey command to generate a public key from a private key
        stored in a slot.

        :param slot: Slot number.
        :type slot: int.
        """
        # Command execution is long. We must wait before querying the response.
        return self.command(ATECCOpCode.GEN_KEY, 0, slot, trigger=trigger, wait=0.2)

    def nonce(self, update_seed=False) -> bytes:
        """
        Call the Nonce command.
        Update `self.temp_key`.
        Warning: this implementation is vulnerable to replay attacks and must be
        used for testing only (num_in is fixed).
        """
        num_in = b"\x55" * 20
        rand_out = self.command(
            ATECCOpCode.NONCE,
            0b01 if update_seed else 0b00,
            0,
            num_in,  # Don't update EEPROM seed
            wait=0.02,
        )
        sha = SHA256.new()
        sha.update(rand_out)
        sha.update(num_in)
        sha.update(
            b"\x16\x01\x00" if update_seed else b"\x16\x00\x00"
        )  # OpCode, mode and param
        # 2 LSB
        self.temp_key = cast(bytes, sha.digest())

        return rand_out

    def set_tempkey(self, k):
        """
        Set the TempKey value using the Nonce command in bypass mode.
        """
        return self.command(ATECCOpCode.NONCE, 0b11, 0, k)

    def mac(self, key_id, challenge):
        """
        :param challenge: Input challenge, 32 bytes long.
        :type challenge: bytes
        """
        return self.command(ATECCOpCode.MAC, (1 << 6), key_id)

    def check_mac(self, key_id: int, key: bytes, trigger=None, use_tempkey=True):
        """
        Calls CheckMac command to authenticate with a given key.
        The method `get_serial` and `nonce` must have been called prior to this
        command to get the device serial number and generate the temp key, both
        used in the MAC calculation.

        :param key_id: Key data slot number.
        :param key: Key value.
        """
        sn = self.serial
        assert sn is not None
        other = b"\x00" * 13  # Extra data used in MAC calculation
        sha = SHA256.new()

        mode = 1 if use_tempkey else 0

        # Mode:1 = 1: MAC hash Tempkey
        # Mode:1 = 0: MAC hash Key
        assert len(key) == 32
        sha.update(key)

        # Mode:0 = 1: MAC hash Tempkey
        # Mode:0 = 0: MAC hash Client Challenge
        client_challenge = b"01234567890123456789012345678901"
        assert len(client_challenge) == 32
        sha.update(self.temp_key if use_tempkey else client_challenge)
        last_block = (
            other[0:4]
            + b"\x00" * 8
            + other[4:7]
            + sn[8:9]
            + other[7:11]
            + sn[0:2]
            + other[11:13]
        )
        assert len(last_block) == 24
        sha.update(last_block)
        mac = sha.digest()
        data = client_challenge + mac + other
        assert len(data) == 32 + 32 + 13

        self.command(ATECCOpCode.CHECK_MAC, mode, key_id, data=data, trigger=trigger)

    def gen_dig(self, key_id: int, key: bytes, trigger=None):
        """
        Calls GenDig command to derive a key from a data slot.
        The method `get_serial` and `nonce` must have been called prior to this
        command to get the device serial number and generate the temp key, both
        used in the new temp key calculation.
        `self.temp_key` is updated.

        :param key_id: Key data slot number.
        :param key: Key value.
        """
        sn = self.serial
        assert sn is not None
        assert self.temp_key is not None
        self.command(ATECCOpCode.GEN_DIG, 2, key_id, trigger=trigger)
        data = (
            key
            + b"\x15"
            + bytes([2])
            + key_id.to_bytes(2, "little")
            + sn[8:9]
            + sn[0:2]
            + b"\x00" * 25
            + self.temp_key
        )
        sha = SHA256.new()
        sha.update(data)
        self.temp_key = cast(bytes, sha.digest())

    def aes(
        self,
        encrypt: bool,
        data_in: bytes,
        key_slot: Optional[int] = None,
        do_nonce=True,
        trigger=None,
    ) -> bytes:
        if key_slot is None:
            if do_nonce:
                self.nonce()
            # 0xFFFF is passed to indicate TempKey
            key_param = 0xFFFF
        else:
            key_param = key_slot
        mode = 0 if encrypt else 1
        return self.command(ATECCOpCode.AES, mode, key_param, data_in, trigger=trigger)



"""Core classes and methods for communicating with the Scaffold FPGA."""

from enum import Enum
from typing import Literal, Optional, TYPE_CHECKING, Union
from abc import ABC, abstractmethod
import serial
from packaging.version import parse as parse_version, Version

if TYPE_CHECKING:
    from scaffold import Scaffold


class TimeoutError(Exception):
    """Thrown when a polling read or write command timed out."""

    def __init__(
        self,
        data: Optional[bytes] = None,
        size: Optional[int] = None,
        expected: Optional[int] = None,
    ):
        """
        :param data: The received data until timeout. None if timeout occured
          during a write operation.
        :param size: The number of successfully proceeded bytes.
        :param expected: The expected number of bytes to be proceeded.
        """
        self.data = data
        self.expected = expected
        if data is not None:
            assert size is None
            self.size = len(data)
        else:
            self.size = size

    def __str__(self):
        if self.data is not None:
            if len(self.data):
                h = self.data.hex()
                return f"Read timeout: partially received {len(self.data)} bytes {h}."
            return "Read timeout: no data received."
        return f"Write timeout. Only {self.size}/{self.expected} bytes written."


class Polling:
    """Polling parameters for read or write operations."""

    def __init__(self, address: int, mask: int, value: int):
        if address not in range(0x10000):
            raise ValueError("Polling address out of range")
        if mask not in range(0x100):
            raise ValueError("Polling mask out of range")
        if value not in range(0x100):
            raise ValueError("Polling value out of range")
        self.address = address
        self.mask = mask
        self.value = value


class OperationStatus(Enum):
    """Possible status of a Scaffold bus operation"""

    PENDING = 0
    SENT = 1
    COMPLETED = 2
    TIMEOUT = 3


class Operation(ABC):
    """
    Base for all Scaffold bus operations.
    Defines how operations are encoded for the board and how the responses are decoded.
    Each operation has an execution status to know wether the response has been
    received or not.
    """

    def __init__(self):
        self.bus: Optional["ScaffoldBus"] = None
        self.status = OperationStatus.PENDING

    def wait(self):
        """Wait until this operation has been processed."""
        assert self.bus is not None
        if self.status == OperationStatus.PENDING:
            raise RuntimeError("Cannot wait a pending operation")
        while self.status == OperationStatus.SENT:
            self.bus.resolve_next_operation()

    @abstractmethod
    def supported(self, version: Version) -> bool:
        """:return: True if the given hardware version supports this operation."""

    @abstractmethod
    def datagram(self) -> bytes:
        """:return: Datagram of the operation, to be sent to the board."""

    @abstractmethod
    def response_size(self) -> int:
        """:return: Expected board response size to this operation."""

    @abstractmethod
    def resolve(self, data: bytes):
        """Process board response."""


class ReadWriteOperation(Operation):
    """Base class for :class:`ReadOperation` and :class:`WriteOperation`."""

    def __init__(self, addr: int, poll: Optional[Polling] = None):
        super().__init__()
        if addr not in range(0x10000):
            raise ValueError("Invalid address")
        self.__addr = addr
        self.__poll = poll

    def supported(self, _version: Version) -> bool:
        return True

    def datagram_head(self, rw: int, size: int) -> bytes:
        """
        Generates part of the datagram for a read write operation.

        :param rw: 0 for read, 1 for write.
        :param size: Read or write size.
        """
        command = rw
        if size > 1:
            command |= 2
        if self.__poll is not None:
            command |= 4
        datagram = bytearray([command, self.__addr >> 8, self.__addr & 0xFF])
        if self.__poll is not None:
            datagram += bytes(
                [
                    self.__poll.address >> 8,
                    self.__poll.address & 0xFF,
                    self.__poll.mask,
                    self.__poll.value,
                ]
            )
        if size > 1:
            datagram.append(size)
        return datagram


class ReadOperation(ReadWriteOperation):
    """Scaffold register read operation"""

    def __init__(self, addr: int, size: int = 1, poll: Optional[Polling] = None):
        super().__init__(addr, poll)
        self.__size = size
        self.__result = bytes()

    def datagram(self) -> bytes:
        return bytes(self.datagram_head(0, self.__size))

    def response_size(self) -> int:
        return 1 + self.__size

    def resolve(self, data: bytes):
        size_completed = data[-1]
        self.__result = data[:size_completed]
        if size_completed == self.__size:
            self.status = OperationStatus.COMPLETED
        else:
            assert size_completed < self.__size
            self.status = OperationStatus.TIMEOUT

    @property
    def result(self) -> bytes:
        """
        Data returned by the board during the read operation.

        :raises TimeoutError: if the read operation timed out.
        """
        self.wait()
        if self.status == OperationStatus.COMPLETED:
            return self.__result
        if self.status == OperationStatus.TIMEOUT:
            raise TimeoutError(data=self.__result, expected=self.__size)
        raise RuntimeError("Invalid operation status")


class WriteOperation(ReadWriteOperation):
    """Scaffold register write operation"""

    def __init__(self, addr: int, data: bytes, poll: Optional[Polling] = None):
        super().__init__(addr, poll)
        if len(data) == 0:
            raise ValueError("No data")
        if len(data) > 255:
            raise ValueError("Data too long")
        self.__data = data

    def datagram(self) -> bytes:
        return bytes(self.datagram_head(1, len(self.__data))) + self.__data

    def response_size(self) -> int:
        return 1

    def resolve(self, data: bytes):
        size_completed = data[0]
        if size_completed == len(self.__data):
            self.status = OperationStatus.COMPLETED
        else:
            assert size_completed < len(self.__data)
            self.status = OperationStatus.TIMEOUT


class TimeoutOperation(Operation):
    """Scaffold timeout configuration operation"""

    def __init__(self, value: int):
        super().__init__()
        if value not in range(0x100000000):
            raise ValueError("Timeout value out of range")
        self.__value = value

    def supported(self, _version: Version) -> bool:
        return True

    def datagram(self) -> bytes:
        return b"\x08" + self.__value.to_bytes(4, "big")

    def response_size(self) -> int:
        return 0

    def resolve(self, data: bytes):
        self.status = OperationStatus.COMPLETED


class DelayOperation(Operation):
    """
    Scaffold delay operation.

    Supported only with hardware >= 0.9.
    """

    def __init__(self, cycles: int):
        super().__init__()
        if cycles not in range(0x1000000):
            raise ValueError("Delay out of range")
        self.__cycles = cycles

    def supported(self, version: Version) -> bool:
        return version >= parse_version("0.9")

    def datagram(self) -> bytes:
        return b"\x09" + self.__cycles.to_bytes(3, "big")

    def response_size(self) -> int:
        return 1

    def resolve(self, data: bytes):
        assert data[0] == 0
        self.status = OperationStatus.COMPLETED


class BufferWaitOperation(Operation):
    """
    Scaffold buffer wait operation.

    Supported only with hardware >= 0.9.
    """

    def __init__(self, size: int):
        super().__init__()
        if size not in range(512):
            raise ValueError("Buffer size out of range")
        self.__size = size

    def supported(self, version: Version) -> bool:
        return version >= parse_version("0.9")

    def datagram(self) -> bytes:
        return b"\x0a" + self.__size.to_bytes(2, "big")

    def response_size(self) -> int:
        return 1

    def resolve(self, data: bytes):
        assert data[0] == 0
        self.status = OperationStatus.COMPLETED


class ScaffoldBusTimeoutSection:
    """
    Helper class to be sure a pushed timeout configuration is poped at some
    time. This is to be used with the python 'with' statement.
    """

    def __init__(self, bus: "ScaffoldBus", timeout: Optional[float]):
        """
        :param bus: Scaffold bus manager.
        :type bus: ScaffoldBus
        :param timeout: Section timeout value, in seconds.
        :type timeout: float, None
        """
        self.bus = bus
        self.timeout = timeout

    def __enter__(self):
        self.bus.push_timeout(self.timeout)

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore
        self.bus.pop_timeout()


class BufferWaitSection:
    """Helper class to enter and leave buffer wait sections."""

    def __init__(self, bus: "ScaffoldBus"):
        self.bus = bus

    def __enter__(self):
        self.bus.push_buffer_wait()

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore
        self.bus.pop_buffer_wait()


class ScaffoldBus:
    """
    Low level methods to drive the Scaffold device.
    """

    MAX_CHUNK = 255
    FIFO_SIZE = 512

    def __init__(self, sys_freq: int, baudrate: int):
        """
        :param baudrate: UART baudrate
        :type baudrate: int
        """
        self.__baudrate = baudrate
        self.sys_freq = sys_freq
        # How long in seconds one timeout unit is.
        self.timeout_unit = 3.0 / self.sys_freq
        self.ser: Optional[serial.Serial] = None
        # Timeout value. This value can't be read from the board, so we cache
        # it there once set.
        self.__cache_timeout = None
        # Timeout stack for push_timeout and pop_timeout methods.
        self.__timeout_stack: list[Optional[float]] = []
        # List of operations that has been sent to the board, and whose response has
        # not been fetched yet.
        self.__operations: list[Operation] = []
        # List of operations that will be sent to the board, prepended with a buffer
        # wait operation to be sure that the board has all the data before starting to
        # execute the operations. This is used to execute different operations with
        # precise timing.
        self.__buffer_wait_operations: list[Operation] = []
        # Every time a buffer wait section is entered, this member is incremented.
        # Every time a buffer wait section is exited, this member is decremented.
        # When this value is greater than zero, operations are not sent to the board,
        # but pushed in the __buffer_wait_operations list. As soon as this value goes
        # back to zero, a buffer wait operation is sent to the board, followed by all
        # pending operations in the list.
        self.__buffer_wait_stack = 0
        self.__fifo_size = 0
        self.version: Optional[Version] = None

    def wait(self):
        """Wait for all pending operations to be completed."""
        while len(self.__operations) > 0:
            self.resolve_next_operation()

    def connect(self, dev: str):
        """
        Connect to Scaffold board using the given serial port.
        :param dev: Serial port device path. For instance '/dev/ttyUSB0' on
            linux, 'COM0' on Windows.
        """
        self.ser = serial.Serial(dev, self.__baudrate)

    def resolve_next_operation(self):
        """
        Reads the response from the oldest :class:`Operation` in the pipe. This will
        block until it has been processed by the board.
        """
        op = self.__operations[0]
        assert self.ser is not None
        op.resolve(self.ser.read(op.response_size()))
        # Remove operation from the queue.
        # If operation has not been copied, this will trigger checks and raise
        # TimeoutError if the operation timed out.
        self.__fifo_size -= len(op.datagram())
        del self.__operations[0]
        if len(self.__operations) == 0:
            assert self.__fifo_size == 0

    def __require_fifo_space(self, size: int):
        """
        Wait for pending operations to complete until there is `size` bytes
        available in the FIFO.
        """
        while self.FIFO_SIZE - self.__fifo_size < size:
            self.resolve_next_operation()

    def operation(self, op: Operation) -> Operation:
        """
        Register a bus operation to be performed.
        """
        if not self.is_connected or self.ser is None:
            raise RuntimeError("Not connected to board")
        if self.version is not None:
            if not op.supported(self.version):
                raise RuntimeError(
                    "Operation not supported by current hardware version"
                )
        if self.__buffer_wait_stack == 0:
            datagram = op.datagram()
            op.bus = self
            assert len(datagram) < self.FIFO_SIZE
            self.__require_fifo_space(len(datagram))
            self.ser.write(datagram)
            op.status = OperationStatus.SENT
            self.__operations.append(op)
            self.__fifo_size += len(datagram)
        else:
            # A synchronized block is being prepared. Don't send the operation to the
            # board yet, keep it in a separate list.
            self.__buffer_wait_operations.append(op)
        return op

    def delay(self, cycles: int):
        """
        Execute a delay operation.

        :param cycles: Number of clock cycles for the delay.
        """
        self.operation(DelayOperation(cycles))

    def write(self, addr: int, data: Union[bytes, int], poll: Optional[Polling] = None):
        """
        Write data to a register.

        :param addr: Register address.
        :param data: Data to be written. Can be a byte, bytes or bytearray.
        :param poll: Register polling parameters, or None if polling is not required.
        """
        # If data is an int, convert it to bytes.
        if isinstance(data, int):
            data = bytes([data])

        offset = 0
        remaining = len(data)
        while remaining:
            chunk_size = min(self.MAX_CHUNK, remaining)
            op = WriteOperation(addr, data[offset : offset + chunk_size], poll)
            self.operation(op)
            remaining -= chunk_size
            offset += chunk_size

    def read(self, addr: int, size: int = 1, poll: Optional[Polling] = None) -> bytes:
        """
        Read data from a register.

        :param addr: Register address.
        :param poll: Register polling parameters, or None if polling is not required.
        """
        result = bytearray()
        remaining = size
        offset = 0
        while remaining:
            chunk_size = min(self.MAX_CHUNK, remaining)
            op = ReadOperation(addr, chunk_size, poll)
            self.operation(op)
            result += op.result
            remaining -= chunk_size
            offset += chunk_size
        return result

    @property
    def is_connected(self):
        """:return: True if connection with a board is established."""
        return self.ser is not None

    @property
    def timeout(self) -> Optional[float]:
        """
        Timeout in seconds for read and write commands. If set to None, timeout
        is disabled.
        """
        if self.__cache_timeout is None:
            raise RuntimeError("Timeout not set yet")
        if self.__cache_timeout == 0:
            return None
        return self.__cache_timeout * self.timeout_unit

    @timeout.setter
    def timeout(self, value: Optional[float]):
        if value is None:
            n = 0
        else:
            n = max(1, int(value / self.timeout_unit))
        if n != self.__cache_timeout:
            self.operation(TimeoutOperation(n))
            self.__cache_timeout = n  # Must be after set_timeout

    def push_timeout(self, value: Optional[float]):
        """
        Save previous timeout setting in a stack, and set a new timeout value.
        Call to `pop_timeout` will restore previous timeout value.
        The new effective timeout will be lower or equal to the current
        timeout. That is, the timeout cannot be increased, previous defined
        timeout have higher priority.

        :param value: New timeout value, in seconds.
        """
        if value is None:
            value = self.timeout
        else:
            if self.timeout is not None:
                value = min(self.timeout, value)
        self.__timeout_stack.append(self.timeout)
        self.timeout = value

    def pop_timeout(self):
        """
        Restore timeout setting from stack.

        :raises RuntimeError: if timeout stack is already empty.
        """
        if len(self.__timeout_stack) == 0:
            raise RuntimeError("Timeout setting stack is empty")
        self.timeout = self.__timeout_stack.pop()

    def timeout_section(self, timeout: Optional[float]) -> ScaffoldBusTimeoutSection:
        """
        :return: :class:`ScaffoldBusTimeoutSection` to be used with the python
            'with' statement to start and close a timeout section.
        """
        return ScaffoldBusTimeoutSection(self, timeout)

    def push_buffer_wait(self):
        """
        Enter a buffer wait section. All next commands won't be sent to the board
        immediately, but will be sent and executed in a single shot when leaving
        the section.
        """
        self.__buffer_wait_stack += 1

    def pop_buffer_wait(self):
        """
        Leave a buffer wait section.
        """
        if self.__buffer_wait_stack == 0:
            raise RuntimeError("No buffer wait section has been started")
        self.__buffer_wait_stack -= 1
        if self.__buffer_wait_stack == 0:
            # Calculate the size of all pending operations
            fifo_size = 0
            for op in self.__buffer_wait_operations:
                fifo_size += len(op.datagram())
            required_fifo_space = fifo_size + 3  # Size of the buffer wait operation
            if required_fifo_space > self.FIFO_SIZE:
                raise RuntimeError("Buffer wait section is too large")
            self.__require_fifo_space(required_fifo_space)
            self.operation(BufferWaitOperation(fifo_size))
            for op in self.__buffer_wait_operations:
                self.operation(op)
            self.__buffer_wait_operations.clear()

    def buffer_wait_section(self) -> BufferWaitSection:
        """
        :return: :class:`BufferWaitSection` to be used with the python 'with' statement
            to start and close a buffer wait section.
        """
        return BufferWaitSection(self)


class Register:
    """
    Manages accesses to a register of a module. Implements value cache
    mechanism whenever possible.
    """

    def __init__(
        self,
        parent: "Scaffold",
        mode: Literal["r", "w", "v"],
        address: int,
        wideness: int = 1,
        min_value: Optional[int] = None,
        max_value: Optional[int] = None,
        reset: Optional[int] = None,
    ):
        """
        :param parent: The Scaffold instance owning the register.
        :param mode: Access mode string. Can have the following characters: 'r'
            for read, 'w' for write, 'v' to indicate the register is volatile.
            When the register is not volatile, a cache is used for read
            accesses.
        :param address: 16-bits address of the register.
        :param wideness: Number of bytes stored by the register. When this
            value is not 1, the register cannot be read.
        :parma min_value: Minimum allowed value. If None, minimum value will be
            0 by default.
        :param max_value: Maximum allowed value. If None, maximum value will be
            2^(wideness*8)-1 by default.
        :param reset: Value to be set to the register when :meth:`reset` is
            called. If None, :meth:`reset` has no effect.
        """
        self.__parent = parent

        if address not in range(0x10000):
            raise ValueError("Invalid register address")
        self.__address = address

        self.__w = "w" in mode
        self.__r = "r" in mode
        self.__volatile = "v" in mode

        if wideness < 1:
            raise ValueError("Invalid wideness")
        if (wideness > 1) and self.__r:
            raise ValueError("Wideness must be 1 if register can be read.")
        self.__wideness = wideness

        if min_value is None:
            # Set default minimum value to 0.
            self.__min_value = 0
        else:
            # Check minimum value.
            if min_value not in range(2 ** (wideness * 8)):
                raise ValueError("Invalid register minimum value")
            self.__min_value = min_value

        if max_value is None:
            # Set default maximum value based on register size.
            self.__max_value = 2 ** (wideness * 8) - 1
        else:
            # Check maximum value.
            if max_value not in range(2 ** (wideness * 8)):
                raise ValueError("Invalid register maximum value")
            self.__max_value = max_value

        if self.__min_value > self.__max_value:
            raise ValueError(
                "Register minimum value must be lower or equal to maximum value"
            )

        self.__reset = reset
        self.__cache = None

    def set(self, value: int, poll: Optional[Polling] = None):
        """
        Set a new value to the register. This method will check bounds against
        the minimum and maximum allowed values of the register. If polling is
        enabled and the register is wide, polling is applied for each byte of
        the register.

        :param value: New value.
        :param poll: Register polling parameters, or None if polling is not required.
        """
        if value < self.__min_value:
            raise ValueError("Value too low")
        if value > self.__max_value:
            raise ValueError("Value too high")
        if not self.__w:
            raise RuntimeError("Register cannot be written")
        # Handle wideness
        value_bytes = value.to_bytes(self.__wideness, "big", signed=False)
        self.__parent.bus.write(self.__address, value_bytes, poll)
        # Save as int
        self.__cache = value

    def get(self) -> int:
        """
        :return: Current register value.
        If the register is not volatile and the value has been cached, no
        access to the board is performed and the cache is returned. If the
        register is not volatile but can't be read, the cached value is
        returned or an exception is raised if cache is not set.
        """
        if self.__volatile:
            if not self.__r:
                raise RuntimeError("Register cannot be read")
            return self.__parent.bus.read(self.__address)[0]
        # Register is not volatile, so its data can be cached.
        if self.__cache is None:
            if self.__r:
                value = self.__parent.bus.read(self.__address)[0]
                self.__cache = value
            else:
                raise RuntimeError("Register cannot be read")
        return self.__cache

    def or_set(self, value: int):
        """
        Sets some bits to 1 in the register.

        :param value: An int.
        """
        self.set(self.get() | value)

    def set_bit(
        self,
        index: int,
        value: Union[Literal[0, 1, True, False], int],
        poll: Optional[Polling] = None,
    ):
        """
        Sets the value of a single bit of the register.

        :param index: Bit index, in [0, 7].
        :param value: True, False, 0 or 1.
        :param poll: Register polling parameters, or None if polling is not required.
        """
        self.set((self.get() & ~(1 << index)) | (int(bool(value)) << index), poll)

    def get_bit(self, index: int) -> int:
        """
        Gets the value of a single bit of the register.

        :return: Value of a given bit, 0 or 1.
        :param index: Bit index, in [0, 7].
        """
        return (self.get() >> index) & 1

    def set_mask(self, value: int, mask: int, poll: Optional[Polling] = None):
        """
        Set selected bits value.

        :param value: Bits value.
        :param mask: A mask indicating which bits must be sets.
        :param poll: Register polling parameters, or None if polling is not required.
        """
        # TODO: raise an exception is the register is declared as volatile ?
        current = self.get()
        self.set((current & (~mask)) | (value & mask), poll)

    def write(self, data: Union[bytes, int], poll: Optional[Polling] = None):
        """
        Raw write in the register. This method raises a RuntimeError if the
        register cannot be written.

        :param data: Data to be written. Can be a int, bytes or bytearray.
        :param poll: Register polling parameters, or None if polling is not required.
        """
        if not self.__w:
            raise RuntimeError("Register cannot be written")
        self.__parent.bus.write(self.__address, data, poll)

    def read(self, size: int = 1, poll: Optional[Polling] = None) -> bytes:
        """
        Raw read the register. This method raises a RuntimeError if the
        register cannot be read.

        :param poll: Register polling parameters, or None if polling is not required.
        """
        if not self.__r:
            raise RuntimeError("Register cannot be read")
        return self.__parent.bus.read(self.__address, size, poll)

    def reset(self):
        """
        Set the register value to its default value. If no default value has
        been defined, this method has no effect.
        """
        if self.__reset is not None:
            self.set(self.__reset)

    def poll(self, mask: int, value: int) -> Polling:
        """
        Creates polling parameters for this register.

        :param value: Expected value for the masked register value.
        :param mask: Mask applied to the register value before comparison with `value`.
        """
        return Polling(self.__address, mask, value)

    @property
    def address(self) -> int:
        """:return: Register address."""
        return self.__address

    @property
    def max(self) -> int:
        """
        Maximum possible value for the register.
        :type: int
        """
        return self.__max_value

    @property
    def min(self) -> int:
        """
        Minimum possible value for the register.
        :type: int
        """
        return self.__min_value



# This file is part of Scaffold
#
# Scaffold is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Copyright 2019 Ledger SAS, written by Olivier Hériveaux


from enum import Enum
from scaffold import Pull, Scaffold, ISO7816
from typing import Callable, Union, Optional, cast
from packaging.version import parse as parse_version
import requests
import crcmod
from time import sleep
from abc import ABC, abstractmethod


class ProtocolError(Exception):
    """
    Exception raised when a protocol error between the terminal and the
    smartcard occurs.
    """

    def __init__(self, *message: object):
        super().__init__(*message)


class Convention(int, Enum):
    """
    Possible ISO-7816 communication convention. This is given by the first byte
    of the ATR returned by the card.
    """

    INVERSE = 0x3F
    DIRECT = 0x3B


class T1RedundancyCode(Enum):
    """
    Possible ISO78-16 possible error detection codes that can be used for T=1
    protocol. It is indicated in the first TC byte for T=1 of the ATR.
    """

    LRC = 0
    CRC = 1


class T1RedundancyCodeError(Exception):
    """Thrown when the redundancy code of a received block is invalid."""

    pass


def inverse_byte(byte: int) -> int:
    """
    Inverse order and polarity of bits in a byte. Used for ISO-7816 inverse
    convention decoding.
    """
    byte ^= 0xFF
    return int(f"{byte:08b}"[::-1], 2)


def apply_convention(data: bytes, convention: Convention) -> bytes:
    """
    :return: `data` if convention is `DIRECT`, `data` with all bytes inversed
        if convention is `INVERSE`.
    """
    if convention == Convention.DIRECT:
        return data
    elif convention == Convention.INVERSE:
        return bytes(inverse_byte(b) for b in data)
    else:
        raise ValueError("invalid convention")


class ATRInfo:
    def __init__(self):
        self.atr = bytearray()
        self.convention = Convention.DIRECT
        self.protocols: set[int] = set()
        self.t_abcd_n: list[list[Optional[int]]] = []


class ByteReader(ABC):
    """Generic class for a reader"""

    @abstractmethod
    def read(self, n: int) -> bytes: ...


class ScaffoldISO7816ByteReader(ByteReader):
    """Used for `parse_atr` function with a Scaffold device."""

    def __init__(self, iso7816: ISO7816):
        """:param iso7816: Scaffold ISO7816 module."""
        self.iso7816 = iso7816
        self.convention = Convention.DIRECT

    def read(self, n: int) -> bytes:
        return apply_convention(self.iso7816.receive(n), self.convention)


class BasicByteReader(ByteReader):
    """Used for `parse_atr` function with a test vector."""

    def __init__(self, data: bytes):
        self.data = data

    def read(self, n: int) -> bytes:
        if len(self.data) < n:
            raise EOFError(
                f"Not enough data: expected {n} bytes "
                f"but got {repr(self.data.hex())} ({len(self.data)} bytes)"
            )
        chunk = self.data[:n]
        self.data = self.data[n:]
        return chunk


def parse_atr(reader: ByteReader) -> ATRInfo:
    """
    ATR parsing function used by Smartcard class when reading a card ATR. The
    reader object passed in arguments allows unit testing with a list of ATR.

    :param reader: An object giving requested bytes for parsing ATR.
    :return: ATR info.
    """
    info = ATRInfo()
    atr = info.atr
    # Receive and parse TS
    atr.append(reader.read(1)[0])
    ts = atr[0]
    try:
        info.convention = Convention(ts)
    except ValueError as e:
        raise ProtocolError(f"Invalid TS byte in ATR: 0x{ts:02x}") from e
    if isinstance(reader, ScaffoldISO7816ByteReader):
        reader.convention = info.convention
    # Receive T0
    atr += reader.read(1)
    # Parse the rest of the ATR
    i = 1
    td = atr[i]
    while td is not None:
        has_t_abcd = list(bool(td & (1 << (j + 4))) for j in range(4))
        count = has_t_abcd.count(True)
        atr += reader.read(count)
        t_abcd: list[Optional[int]] = [None, None, None, None]
        offset = 0
        for j in range(4):
            if has_t_abcd[j]:
                t_abcd[j] = atr[-count + offset]
                offset += 1
        info.t_abcd_n.append(t_abcd)
        # Test to skip T0 byte
        if i != 1:
            info.protocols.add(td & 0x0F)
        # Test TD presence
        if has_t_abcd[3]:
            td = atr[-1]
        else:
            td = None
        i += 1
    # If no protocol is specified, then T=0 is available by default
    if len(info.protocols) == 0:
        info.protocols.add(0)
    # Fetch historical bytes
    # Number of historical bytes is the low nibble of T0
    num_historical_bytes = atr[1] & 0x0F
    try:
        atr += reader.read(num_historical_bytes)
    except EOFError as e:
        raise EOFError(
            f"Not enough data for historical bytes in {atr.hex()}:"
            + f" expected {num_historical_bytes}"
            + (
                f", got {len(reader.data)}"
                if isinstance(reader, BasicByteReader)
                else ""
            )
            + f": {str(e)}"
        )

    # Parse TCK (check byte)
    # This byte is absent if only T=0 is supported
    if info.protocols != {0}:
        # TCK expected
        atr += reader.read(1)
        # tck = atr[-1]
        # Verify the checksum
        xored = 0x00
        for b in atr[1:]:
            xored ^= b
        if xored != 0x00:
            raise ProtocolError("ATR checksum error")
    return info


class NoATRDatabase(Exception):
    """
    Thrown during card information lookup if no ATR database could be loaded
    """

    pass


def load_atr_info_db(allow_web_download: bool = False) -> list[tuple[str, list[str]]]:
    """
    Parse the smartcard ATR list database from Ludovic Rousseau to get list of
    known ATR.

    The database file cannot be embedded in the library because it uses GPL
    and not LGPL license. On debian systems, this file is provided in the
    pcsc-tools package. If this file is missing and `allow_web_download` is
    enabled, this method will read the file with an HTTP GET request to:
    http://ludovic.rousseau.free.fr/softwares/pcsc-tools/smartcard_list.txt

    ATR values are returned with strings, and can have '.' wildcards for
    matching, or other special formatting characters. With each ATR is returned
    of list of description strings.

    :param allow_web_download: If enabled, allow the method to download the
        database from the web as a fallback when it is missing from the system.
    :raises NoATRDatabase: When database file is missing and download is not
        allowed, or when database file is missing and download failed.
    """
    tab: list[tuple[str, list[str]]] = []
    try:
        text_file = open("/usr/share/pcsc/smartcard_list.txt", "r")
        # We don't want to keep end lines such as LR or CR LF
        lines = text_file.read().splitlines()
        text_file.close()
    except FileNotFoundError:
        if allow_web_download:
            url = (
                "http://ludovic.rousseau.free.fr/softwares/pcsc-tools/"
                + "smartcard_list.txt"
            )
            try:
                res = requests.get(url)
            except Exception:
                raise NoATRDatabase()
            if res.status_code != 200:
                raise NoATRDatabase()
            lines = res.content.decode().splitlines()
        else:
            raise NoATRDatabase()
    # Parse the file and build a table with ATR patterns and infos
    for line in lines:
        if (len(line) > 0) and (line[0] not in ("#", "\t")):
            # ATR line
            atr = line.replace(" ", "").lower()
            tab.append((atr, []))
        elif (len(line) > 0) and (line[0] == "\t"):
            # Info line
            # Remove first character \t
            tab[-1][1].append(line[1:])
    return tab


class Smartcard:
    """
    Class for smartcard testing with Scaffold board and API. The
    following IOs are used:

    - D0: ISO-7816 IO
    - D1: ISO-7816 nRST
    - D2: ISO-7816 CLK
    - D3: Socket card contactor sense

    :class:`scaffold.Scaffold` class has ISO-7816 peripheral support, but it is
    very limited. This class enables full support to ISO-7816 by managing ATR,
    convention convertion, T=0 or T=1 protocols, etc.

    :var bytes atr: ATR received from card after reset.
    :var Convention convention: Communication convention between card
        and terminal. Updated when first byte TS of ATR is received.
    :var set protocols: Communication protocols found in ATR. This set contains
        integers, for instance 0 if T=0 is supported, 1 if T=1 is supported...
    """

    def __init__(self, scaffold: Optional[Scaffold] = None):
        """
        Configure a Scaffold board for use with smartcards.

        :param scaffold: Board instance which will be configured as a smartcard
            reader. :class:`scaffold.Scaffold` instance.
        """
        if scaffold is None:
            scaffold = Scaffold()
        self.iso7816 = scaffold.iso7816
        self.scaffold = scaffold
        self.sig_nrst = scaffold.d1
        self.sig_sense = scaffold.d3
        _ = self.sig_nrst << 1
        # D0 is connected to a bidirectionnal bus. We enable the pull-up
        # resistor if hardware version is >= 1.1. For version 1.0, the pull-up
        # resistor must be soldered on the daughterboard.
        # 1.0 hardware version boards have <= 0.3 architecture version.
        assert scaffold.version is not None
        if scaffold.version >= parse_version("0.3"):
            scaffold.d0.pull = Pull.UP
        _ = scaffold.d0 << scaffold.iso7816.io_out
        _ = scaffold.d0 >> scaffold.iso7816.io_in
        _ = scaffold.d2 << scaffold.iso7816.clk
        self.atr = bytearray()
        self.convention = Convention.DIRECT
        self.crc16: Optional[Callable[[bytearray], int]] = None
        self.t1_ns_tx = 0  # Sequence number for I-block tranmission in T=1
        self.t1_ns_rx = 0  # Sequence number for I-block reception in T=1

    def receive(self, n: int) -> bytes:
        """
        Use the ISO-7816 peripheral to receive bytes from the smartcard, and
        apply direct or inverse convention depending on what has been read in
        the ATR.

        :param n: Number of bytes to be read.
        """
        return apply_convention(self.iso7816.receive(n), self.convention)

    def reset(self) -> bytes:
        """
        Reset the smartcard and retrieve the ATR.

        If the ATR is retrieved successfully, the attributes :attr:`atr`
        :attr:`convention` and :attr:`protocols` are updated.

        :attr:`protocols` indicates what protocols are supported. It will
            contain 0 if T=0 is supported, and 1 if T=1 is supported.

        If only T=1 is supported, exchanges using :meth:`apdu` will use I-block
        transmission automatically.

        :return: ATR from the card.
        :raises ProtocolError: if the ATR is not valid.
        """
        _ = self.sig_nrst << 0
        sleep(0.05)
        self.iso7816.flush()
        _ = self.sig_nrst << 1
        info = parse_atr(ScaffoldISO7816ByteReader(self.iso7816))
        self.atr = info.atr
        self.convention = info.convention
        self.protocols = info.protocols
        self.atr_info = info
        # If T=1 is supported, read in TC1 the correct redundancy code to be
        # used
        if 1 in self.protocols:
            self.t1_ns_tx = 0
            self.t1_ns_rx = 0

            # LRC is default value
            self.t1_redundancy_code = T1RedundancyCode.LRC
            for i in range(2, len(self.atr_info.t_abcd_n)):
                tci = self.atr_info.t_abcd_n[i][2]
                tdi = self.atr_info.t_abcd_n[i][3]
                if tdi is not None and (tdi & 0x0F) == 1 and tci is not None:
                    # We found a byte specific for T=1 protocol
                    self.t1_redundancy_code = T1RedundancyCode(tci & 1)
                    break
        else:
            self.t1_redundancy_code = None
        # Verify that there are no more bytes
        if not self.iso7816.empty:
            raise ProtocolError("Unexpected bytes after ATR")
        return info.atr

    def apdu(self, the_apdu: Union[bytes, str], trigger: str = "") -> bytes:
        """
        Send an APDU to the smartcard and retrieve the response.

        If only T=1 protocol is supported, this method use `transmit_block` and
        `receive_block` to send the APDU by sending information blocks.

        :param the_apdu: APDU to be sent. str hexadecimal strings are allowed,
            but user should consider using the :meth:`apdu_str` method instead.
        :param trigger: If 'a' is in this string, trigger is raised after
            ISO-7816 header is sent in T=0, and cleared when the following
            response byte arrives. If 'b' is in this string, trigger is raised
            after data field has been transmitted in T=0, and cleared when the
            next response byte is received. If T=1, both 'a' and 'b' will raise
            trigger after the I-block has been transmitted, and will be cleared
            when first byte of next block is received.
        :raises ValueError: if APDU data is invalid.
        :raises RuntimeError: if the received procedure byte is invalid in T=0,
            or if neither T=0 and T=1 protocols are supported by the card.
        :raises T1RedundancyCodeError: If LRC or CRC is wrong in T=1 protocol.
        :return: Response data, with status word.
        """
        if isinstance(the_apdu, str):
            the_apdu = bytes.fromhex(the_apdu)
        the_apdu = cast(bytes, the_apdu)

        apdu_len = len(the_apdu)
        if apdu_len < 5:
            raise ValueError("APDU too short")
        if 0 in self.protocols:
            return self.__apdu_t0(the_apdu, trigger)
        elif 1 in self.protocols:
            return self.__apdu_t1(the_apdu, ("a" in trigger) or ("b" in trigger))
        else:
            raise RuntimeError("Neither T=0 or T=1 are supported by the card")

    def __apdu_t0(self, the_apdu: bytes, trigger: str = "") -> bytes:
        """
        Send an APDU to the smartcard and retrieve the response, using T=0
        protocol.

        :param the_apdu: APDU to be sent.
        :param trigger: If 'a' is in this string, trigger is raised after
            ISO-7816 header is sent, and cleared when the following response
            byte arrives. If 'b' is in this string, trigger is raised after
            data field has been transmitted, and cleared when the next
            response byte is received.
        :raises ValueError: if APDU data is invalid.
        :raises RuntimeError: if the received procedure byte is invalid.
        :return: Response data, with status word.
        """
        out_data_len = len(the_apdu) - 5
        if out_data_len > 256:
            raise ValueError("APDU too long")
        if out_data_len > 0:
            # This is an outgoing data transfer
            # Verify APDU P3 field correctness
            p3 = the_apdu[4]
            expected_p3 = out_data_len % 256
            if p3 != expected_p3:
                raise ValueError(
                    "Expected P3 (length) in APDU is "
                    f"0x{expected_p3:02x}, got 0x{p3:02x}"
                )
            in_data_len = 0
        else:
            if the_apdu[4] > 0:
                in_data_len = the_apdu[4]
            else:
                in_data_len = 256
        # Transmit the header
        if "a" in trigger:
            self.iso7816.transmit(the_apdu[:4])
            self.iso7816.trigger_long = True
            self.iso7816.transmit(the_apdu[4:5])
        else:
            # Send all the header at once
            self.iso7816.trigger_long = False
            self.iso7816.transmit(the_apdu[:5])
        # Receive procedure byte
        procedure_byte = self.iso7816.receive(1)[0]
        if "a" in trigger:  # Disable only if enabled previously
            self.iso7816.trigger_long = False
        while procedure_byte == 0x60:
            procedure_byte = self.iso7816.receive(1)[0]
        response = bytearray()
        ins = the_apdu[1]
        if (procedure_byte & 0xF0) in (0x60, 0x90):
            # Received SW1 byte.
            response.append(procedure_byte)
            # Received SW2
            response.append(self.iso7816.receive(1)[0])
            return response
        elif procedure_byte in (ins, ~ins):
            # Acknowledge byte.
            # Transfer the remaining data
            if out_data_len > 0:
                if "b" in trigger:
                    # Enable trigger on last byte only
                    self.iso7816.transmit(the_apdu[5:-1])
                    self.iso7816.trigger_long = True
                    self.iso7816.transmit(the_apdu[-1:])
                else:
                    # Send all remaining data at once
                    self.iso7816.transmit(the_apdu[5:])
            # Receive the response data and status word
            response += self.iso7816.receive(in_data_len + 2)
            if "b" in trigger:  # Disable only if enabled previously
                self.iso7816.trigger_long = False
            return response
        else:
            raise RuntimeError(
                f"Unexpected procedure byte 0x{procedure_byte:02x} received"
            )

    def __apdu_t1(self, the_apdu: bytes, trigger: bool = False) -> bytes:
        """
        Send an APDU to the smartcard and retrieve the response, using T=1
        protocol. Transmission supports chaining for both command and response.

        :param the_apdu: APDU to be sent.
        :param trigger: If True, raise trigger on last byte of last I-block
            transmission.
        :return: Response data, with status word.
        :raises T1RedundancyCodeError: If LRC or CRC is wrong.
        :raises ProtocolError: If an unpexpected response is received.
        """
        edc_len_dict = {T1RedundancyCode.LRC: 1, T1RedundancyCode.CRC: 2}
        assert self.t1_redundancy_code is not None

        enable_trigger = trigger

        apdu_remaining = the_apdu
        while len(apdu_remaining):
            chunk = apdu_remaining[:254]
            apdu_remaining = apdu_remaining[254:]
            has_more = min(1, len(apdu_remaining))
            enable_trigger = trigger and not has_more
            # Send I-block
            self.transmit_block(
                0, (self.t1_ns_tx << 6) + (has_more << 5), chunk, enable_trigger
            )
            # Increment sequence number
            self.t1_ns_tx = (self.t1_ns_tx + 1) % 2
            if has_more:
                # We expect a R-block before sending the next info block
                block = self.receive_block()
                pcb = block[1]
                if pcb & (1 << 7) == 0:  # I-block
                    raise ProtocolError("Expected R-block, received I-block")
                elif pcb & (1 << 6) == 0:  # R-block
                    if pcb & 0x0F == 0:  # Error free acknowledgement
                        pass
                    elif pcb & 0x0F == 1:
                        raise ProtocolError("Redundancy code error reported by card")
                    else:
                        raise ProtocolError("Unspecified error reported by card")
                else:  # S-block
                    raise ProtocolError("Expected R-block, received I-block")

        response = bytearray()
        has_more = True
        while has_more:
            block = self.receive_block()
            if enable_trigger:
                self.iso7816.trigger_long = False
            pcb = block[1]
            if pcb & (1 << 7) == 0:  # I-block
                # Check that the sequence number is correct
                if (pcb >> 6) & 1 != self.t1_ns_rx:
                    raise ProtocolError(
                        "Incorrect received sequence number in I-block " + block.hex()
                    )
                has_more = bool((pcb >> 5) & 1)
                edc_len = edc_len_dict[self.t1_redundancy_code]
                response += block[3:-edc_len]  # Trim header and EDC
            elif pcb & (1 << 6) == 0:  # R-block
                raise ProtocolError("Expected I-block, received R-block")
            else:  # S-block
                raise ProtocolError("Expected I-block, received S-block")

            self.t1_ns_rx = (self.t1_ns_rx + 1) % 2
            if has_more:
                # Send R block
                self.transmit_block(0, 0b10000000 + (self.t1_ns_rx << 4), b"")

        return response

    def pps(self, pps1: int) -> int:
        """
        Send a PPS request to change the communication speed parameters Fi and
        Di (as specified in ISO-7816-3). PPS0 and PPS1 are sent. PPS2 is
        ignored. This method waits for the response of the card and then
        automatically changes the ETU from the Fi and Di values.
        Scaffold hardware does not support all possible parameters:
        ETU = Fi/Di must not have a fractional part.

        :param pps1: Value of the PPS1 byte.
        :return: New etu value.
        :raises ValueError: if Fi or Di parameters in PPS1 are reserved.
        :raises ValueError: if target ETU has a fractional part.
        :raises RuntimeError: if response to PPS request is invalid.
        """
        if pps1 not in range(0x100):
            raise ValueError("Invalid PPS1 value")
        fi: Optional[int] = [
            372,
            372,
            558,
            744,
            1116,
            1488,
            1860,
            None,
            None,
            512,
            768,
            1024,
            1536,
            2048,
            None,
            None,
        ][pps1 >> 4]
        if fi is None:
            raise ValueError("Fi parameter in PPS1 has a reserved value")
        di = [None, 1, 2, 4, 8, 16, 32, 64, 12, 20, None, None, None, None, None, None][
            pps1 & 0x0F
        ]
        if di is None:
            raise ValueError("Di parameter in PPS1 has a reserved value")
        # PPSS = 0xff
        # PPS0 = 0x10 to indicate presence of PPS1
        request = bytearray(b"\xff\x10")
        request.append(pps1)
        etu = round(fi / di)
        if etu != fi / di:
            raise ValueError(
                f"Cannot set ETU to {etu} because of the "
                "fractional part (hardware limitation)"
            )
        # Checksum
        pck = 0
        for b in request:
            pck ^= b
        request.append(pck)
        # Send the request
        request = request
        self.iso7816.transmit(request)
        # Get the response
        res = self.iso7816.receive(4, timeout=1)
        if res == request:
            # Negociation is successfull
            self.iso7816.etu = etu
            return etu
        else:
            raise RuntimeError("PPS request failed")

    def find_info(self, allow_web_download: bool = False) -> Optional[list[str]]:
        """
        Parse the smartcard ATR list database available at
        http://ludovic.rousseau.free.fr/softwares/pcsc-tools/smartcard_list.txt
        and try to match the current ATR to retrieve more info about the card.

        The database file cannot be embedded in the library because it uses GPL
        and not LGPL license. On debian systems, this file is provided in the
        pcsc-tools package.

        :param allow_web_download: If enabled, allow the method to download the
            database from the web as a fallback when it is missing from the
            system.
        :return: A list of str, where each item is an information line about
            the card. Return None if the ATR did not match any entry in the
            database.
        :raises NoATRDatabase: When database file is missing and download is
            not allowed, or when database file is missing and download failed.
        """
        tab = load_atr_info_db(allow_web_download)
        # Try to match ATR
        for item in tab:
            pattern = item[0]
            atr = self.atr.hex()
            if len(pattern) != len(atr):
                continue
            match = True
            for i in range(len(pattern)):
                if pattern[i] != ".":
                    if pattern[i] != atr[i]:
                        match = False
                        break
            if match:
                return item[1]
                break

    def apdu_str(self, the_apdu: str) -> str:
        """
        Same as :meth:`apdu` function, with str argument and return type for
        convenience.

        :param the_apdu: APDU to be sent, as an hexadecimal string.
        :return: Response from the card, as a lowercase hexadecimal string
            without spaces.
        """
        return self.apdu(bytes.fromhex(the_apdu)).hex()

    def calculate_edc(self, data: bytes) -> bytes:
        """
        Calculate expected error detection code for a block. Depending on
        `self.t1_redundancy_code` LRC (1 bytes) or CRC (2 bytes) is calculated.

        :param data: Input data for the error detection code calculation.
            Includes all bytes of the block excepted the LRC or CRC bytes.
        :return: ECC bytes at the end of the block.
        """
        if self.t1_redundancy_code == T1RedundancyCode.LRC:
            result = 0
            for b in data:
                result = result ^ b
            return bytes([result])
        elif self.t1_redundancy_code == T1RedundancyCode.CRC:
            # CRC is CRC-16-CCITT, with initial value 0xffff
            # I cannot tell if this is correct as it does not seem very well
            # documented and I don't have any card to test this...
            if self.crc16 is None:
                self.crc16 = crcmod.mkCrcFun(0x11021, 0xFFFF, rev=False)  # type: ignore
            return self.crc16(data).to_bytes(2, "big")  # type: ignore
        else:
            raise RuntimeError("invalid t1_redundancy_code value")

    def transmit_block(
        self, nad: int, pcb: int, info: bytes = b"", trigger: bool = False
    ):
        """
        Transmit a T=1 protocol block.
        Error detection code is calculated and appended automatically.

        :param nad: Node address byte.
        :param pcb: Protocol control byte.
        :param info: Information field.
        :param trigger: If True, raise trigger on last byte transmission.
        """
        if len(info) > 254:
            raise ValueError(f"info field is too long ({len(info)} > 254)")
        data = bytearray([nad, pcb, len(info)]) + info
        data += self.calculate_edc(data)
        if trigger:
            self.iso7816.transmit(data[:-1])
            self.iso7816.trigger_long = True
            self.iso7816.transmit(data[-1:])
        else:
            self.iso7816.trigger_long = False
            self.iso7816.transmit(data)

    def receive_block(self) -> bytes:
        """
        Receive a T=1 protocol block.

        :raises T1RedundancyCodeError: If LRC or CRC is wrong.
        """
        block = self.iso7816.receive(3)
        if self.t1_redundancy_code == T1RedundancyCode.LRC:
            edc_len = 1
        elif self.t1_redundancy_code == T1RedundancyCode.CRC:
            edc_len = 2
        else:
            # This should not happen as we are receiving a T=1 block
            assert False
        block += self.iso7816.receive(block[2] + edc_len)
        # Verify LRC/CRC
        if self.calculate_edc(block[:-edc_len]) != block[-edc_len:]:
            raise T1RedundancyCodeError()
        return block

    @property
    def card_inserted(self):
        """
        True if a card is inserted, False otherwise. Card insertion is detected
        with a mecanical switch connected to D3 of Scaffold.
        """
        return self.sig_sense.value == 1
